from inspect import stack
from logging import Logger, getLogger
from random import choice

import pandas as pd
from dash import html, dcc, callback, Output, Input, State
from dash_extensions import EventListener

# from pandas import DataFrame
from pandas import DataFrame
from plotly.express import scatter
from plotly.graph_objects import Figure

from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.widget import WidgetAsDiv


class BubblesAsDiv(WidgetAsDiv):
    """
    python3 ./src/enov.py --verbose df-load-xls ../2022-11-28.HistoricalPipelineForBubbles.xlsx df_raw dict-load-json ../config.json.txt config ds-dash config

    """
    _logger: Logger = getLogger(__name__)

    # The widget IDs to be used
    # Naming convention:
    # - should start by "const_str__wid__"
    # - as the application will screen all the widget ids, across widgets, to ensure there is no ambiguity, aka same
    #   widget id across different widgets
    const_str__wid__bubbles_as_div: str = "enov-content__bubbles__as_div"
    const_str__wid__bubbles_as_dcc_store__data: str = "enov-content__bubbles__as_dcc_store__data"
    const_str__wid__bubbles_as_dcc_store__filters: str = "enov-content__bubbles__as_dcc_store__filters"
    const_str__wid__bubbles_as_graph: str = "enov-content__bubbles__as_graph"

    const_str__config_label__dict_px_scatter_parameters: str = "px.scatter"
    const_str__config_label__str_dataframe_name_in_context: str = "data"

    """
    Return the bubbles widget such as:
    TODO
    """

    def __init__(
            self,
            dict_config: dict,
            dict_context: dict,
            session: Session
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # ##############################################################################################################
        # Callbacks should be imported here... Otherwise, we face "unknown id" as callbacks get registered even though
        # the widgets are not used...
        # noinspection PyUnresolvedReferences
        # import com.enovation.toolbox.dash.widgets.bubbles.bubbles_callbacks
        # ##############################################################################################################

        # df_data: DataFrame = DataFrame()
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        super().__init__(
            dict_config=dict_config,
            dict_context=dict_context,
            session=session,
            children=[
                dcc.RadioItems(
                    id='selection',
                    options=["GDP - Scatter", "Population - Bar"],
                    value='GDP - Scatter',
                ),
                html.Div(
                    dcc.Loading(
                        dcc.Graph(
                            id=self.const_str__wid__bubbles_as_graph,
                            style={'height': '90vh'}
                        ),
                        type="cube"),
                    id="container"
                ),

                EventListener(
                    id='eventlistener',
                    children=html.Div("Click here!", className="stuff"),
                    events=[
                        {
                            "event": "click", "props": ["srcElement.className"]
                        }
                    ],
                    logging=True
                ),
                html.Div(id="jsg"), html.Div(id="n_events"),

                # dcc.Store for the data to be loaded from the server, and then displayed
                dcc.Store(id=self.const_str__wid__bubbles_as_dcc_store__data),
            ],
            id=BubblesAsDiv.const_str__wid__bubbles_as_div,
            className="container border bg-light",
            style={"margin-top": "1rem", 'height': '90vh'}
        )

        session.dash_application.dash_server.clientside_callback(
            """

            function(n_events, event){
                
                if( typeof event === 'undefined' ) {
                    return "";
                }
                
                if( event["srcElement.className"] == "stuff" ) {

                    // if already full screen; exit
                    // else go fullscreen
                    if (
                        document.fullscreenElement ||
                        document.webkitFullscreenElement ||
                        document.mozFullScreenElement ||
                        document.msFullscreenElement
                    ) {
                        if (document.exitFullscreen) {
                            document.exitFullscreen();
                        } else if (document.mozCancelFullScreen) {
                            document.mozCancelFullScreen();
                        } else if (document.webkitExitFullscreen) {
                            document.webkitExitFullscreen();
                        } else if (document.msExitFullscreen) {
                            document.msExitFullscreen();
                        }
                    } else {
                        element = document.getElementById("container");
                        if (element.requestFullscreen) {
                            element.requestFullscreen();
                        } else if (element.mozRequestFullScreen) {
                            element.mozRequestFullScreen();
                        } else if (element.webkitRequestFullscreen) {
                            element.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
                        } else if (element.msRequestFullscreen) {
                            element.msRequestFullscreen();
                        }
                    }

                    return "ZOOM!!"+n_events+" - "+JSON.stringify(event);
                }

                return "NOTHING!!"+n_events+" - "+JSON.stringify(event);
            }
            """,
            Output("jsg", "children"),
            Input("eventlistener", "n_events"),
            State("eventlistener", "event")
        )

        @callback(
            Output(BubblesAsDiv.const_str__wid__bubbles_as_graph, "figure"),
            Input("selection", "value")
        )
        def display_animated_graph(selection):

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

            _df_the_data: DataFrame = self.session.dict_context.get(
                self.dict_config.get(self.const_str__config_label__str_dataframe_name_in_context)
            )

            jsg = self.jsg()

            _the_timestamps = _df_the_data["Sp - Snapshot Date"].unique()

            _the_opp_ids = _df_the_data["Opp - ID"].unique()


            # We randomly set text position for each opportunity id
            _df_the_opp_ids_and_text_position: DataFrame = DataFrame()
            _df_the_opp_ids_and_text_position["Opp - ID"] = _the_opp_ids
            _df_the_opp_ids_and_text_position["position"] = _df_the_opp_ids_and_text_position.apply(
                lambda row: choice(
                    [
                        'top left',
                        'top center',
                        'top right',
                        'bottom left',
                        'bottom center',
                        'bottom right',
                        'middle left',
                        'middle center',
                        'middle right',
                    ]
                ), axis=1
            )
            _df_the_data = _df_the_data.merge(right=_df_the_opp_ids_and_text_position, how="left", on="Opp - ID")

            # Make the list of BUs
            _the_bus = _df_the_data["Cpy - BU"].unique()

            # make figure
            _dict_the_figure: dict = {
                "data": [],
                "layout": {},
                "frames": []
            }

            # fill in most of layout
            _dict_the_figure["layout"]["xaxis"] = {"range": ["2022-06-01", "2023-05-31"], "title": "kick off"}
            _dict_the_figure["layout"]["yaxis"] = {"range": [-0.2, 1.2], "title": "Confidence level"}
            _dict_the_figure["layout"]["hovermode"] = "closest"
            _dict_the_figure["layout"]["updatemenus"] = [
                {
                    "buttons": [
                        {
                            "args": [None, {"frame": {"duration": 1000, "redraw": False},
                                            "fromcurrent": True, "transition": {"duration": 1000,
                                                                                "easing": "linear"}}],
                            "label": "Play",
                            "method": "animate"
                        },
                        {
                            "args": [[None], {"frame": {"duration": 0, "redraw": False},
                                              "mode": "immediate",
                                              "transition": {"duration": 0}}],
                            "label": "Pause",
                            "method": "animate"
                        }
                    ],
                    "direction": "left",
                    "pad": {"r": 10, "t": 87},
                    "showactive": False,
                    "type": "buttons",
                    "x": 0.1,
                    "xanchor": "right",
                    "y": 0,
                    "yanchor": "top"
                }
            ]

            _dict_the_sliders = {
                "active": 0,
                "yanchor": "top",
                "xanchor": "left",
                "currentvalue": {
                    "font": {"size": 20},
                    "prefix": "Timestamp:",
                    "visible": True,
                    "xanchor": "right"
                },
                "transition": {"duration": 0, "easing": "cubic-in-out"},
                "pad": {"b": 10, "t": 50},
                "len": 0.9,
                "x": 0.1,
                "y": 0,
                "steps": []
            }

            # make data
            _time_the_timestamp = min(_df_the_data["Sp - Snapshot Date"])
            for i_bu in _the_bus:
                _df_the_data_by_timestamp = _df_the_data[_df_the_data["Sp - Snapshot Date"] == _time_the_timestamp]
                _df_the_data_by_timestamp_and_bu = _df_the_data_by_timestamp[
                    _df_the_data_by_timestamp["Cpy - BU"] == i_bu]

                _dict_the_data: dict = {
                    "x": list(_df_the_data_by_timestamp_and_bu["Sp - Kick Off Date"]),
                    "y": list(_df_the_data_by_timestamp_and_bu["Sp - Conf Lvl"]),
                    "mode": "markers+text",
                    "text": list(_df_the_data_by_timestamp_and_bu["Opp - ID"]),
                    "textposition": list(_df_the_data_by_timestamp_and_bu["position"]),
                    "marker": {
                        "sizemode": "area",
                        "sizeref": 1,
                        "size": list(_df_the_data_by_timestamp_and_bu["Sp - Effort"])
                    },
                    "name": i_bu
                }

                _dict_the_figure["data"].append(_dict_the_data)

            _dict_the_figure["data"].append(
                    {
                        "x": [_time_the_timestamp, _time_the_timestamp],
                        "y": [-0.2, 1.2],
                        "mode": "lines",
                        "name": "lines",
                        "ids": ["timestamp1", "timestamp2"]
                    }
            )

            # make frames
            for i_timestamp in _the_timestamps:
                _the_frame = {"data": [], "name": str(i_timestamp)}

                for i_bu in _the_bus:
                    _df_the_data_by_timestamp = _df_the_data[_df_the_data["Sp - Snapshot Date"] == i_timestamp]

                    _jsg = DataFrame()
                    _jsg["Opp - ID"] = _the_opp_ids
                    _jsg = pd.merge(
                        how="left",
                        right=_df_the_data_by_timestamp,
                        left=_jsg,
                        on="Opp - ID"
                    )

                    # _df_the_data_by_timestamp_and_bu = _df_the_data_by_timestamp[
                    #     _df_the_data_by_timestamp["Cpy - BU"] == i_bu]

                    _df_the_data_by_timestamp_and_bu = _jsg[
                        _jsg["Cpy - BU"] == i_bu]

                    _dict_the_data: dict = {
                        "x": list(_df_the_data_by_timestamp_and_bu["Sp - Kick Off Date"]),
                        "y": list(_df_the_data_by_timestamp_and_bu["Sp - Conf Lvl"]),
                        "mode": "markers+text",
                        "text": list(_df_the_data_by_timestamp_and_bu["Opp - ID"]),
                        "textposition": list(_df_the_data_by_timestamp_and_bu["position"]),
                        "marker": {
                            "sizemode": "area",
                            "sizeref": 1,
                            "size": list(_df_the_data_by_timestamp_and_bu["Sp - Effort"])
                        },
                        "name": i_bu,
                        "ids": list(_df_the_data_by_timestamp_and_bu["Opp - ID"])
                    }
                    _the_frame["data"].append(_dict_the_data)

                _the_frame["data"].append(
                    {
                        "x": [i_timestamp, i_timestamp],
                        "y": [-0.2, 1.2],
                        "mode": "lines",
                        "ids": ["timestamp1", "timestamp2"]
                    }
                )

                _dict_the_figure["frames"].append(_the_frame)

                _dict_the_slider_step = {"args": [
                    [str(i_timestamp)],
                    {"frame": {"duration": 1000, "redraw": False},
                     "mode": "immediate",
                     "transition": {"duration": 1000}}
                ],
                    "label": str(i_timestamp)[:10],
                    "method": "animate"}
                _dict_the_sliders["steps"].append(_dict_the_slider_step)

            _dict_the_figure["layout"]["sliders"] = [_dict_the_sliders]

            _figure_the_return: Figure = Figure(_dict_the_figure)

            #_figure_the_return.update_traces(textposition='top center')

            self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

            # return animations[selection]
            return _figure_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    # Version that works... But I cannot
    # @callback(
    #     Output(BubblesAsDiv.const_str__wid__bubbles_as_graph, "figure"),
    #     Input("selection", "value")
    # )
    def jsg(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _df_the_data: DataFrame = self.session.dict_context.get(
            self.dict_config.get(self.const_str__config_label__str_dataframe_name_in_context)
        )

        _fun_parameters: dict = self.dict_config.get(self.const_str__config_label__dict_px_scatter_parameters, {})
        _df_the_data[_fun_parameters["animation_frame"]] = \
            _df_the_data[_fun_parameters["animation_frame"]].astype(str)
        # ["animation_frame"] = _df_the_data[_fun_parameters["animation_frame"]].astype(str)

        if _fun_parameters.get("range_x", "") == "auto":
            _fun_parameters["range_x"] = [
                min(_df_the_data[_fun_parameters["x"]]),
                max(_df_the_data[_fun_parameters["x"]])
            ]

        if _fun_parameters.get("range_y", "") == "auto":
            _fun_parameters["range_y"] = [
                min(_df_the_data[_fun_parameters["y"]]),
                max(_df_the_data[_fun_parameters["y"]])
            ]

        _figure_the_return: Figure = scatter(
            **{
                "data_frame": _df_the_data,
                **_fun_parameters
            }
        )

        _figure_the_return.layout.updatemenus[0].buttons[0].args[1]["frame"]["duration"] = 100

        # df = px.data.gapminder()  # replace with your own data source
        # _local: Session = self.session
        #
        # animations = {
        #     'GDP - Scatter': px.scatter(
        #         df, x="gdpPercap", y="lifeExp", animation_frame="year",
        #         animation_group="country", size="pop", color="continent",
        #         hover_name="country", log_x=True, size_max=55,
        #         range_x=[100, 100000], range_y=[25, 90]),
        #     'Population - Bar': px.bar(
        #         df, x="continent", y="pop", color="continent",
        #         animation_frame="year", animation_group="country",
        #         range_y=[0, 4000000000]),
        # }
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        # return animations[selection]
        return _figure_the_return



from pandas import DataFrame


class PredictabilityBean:

    def __init__(
            self,
            df_historical: DataFrame,
            df_by_measure: DataFrame,
            df_by_key: DataFrame,
    ):
        self._df_historical = df_historical
        self._df_by_measure = df_by_measure
        self._df_by_key = df_by_key

    @property
    def df_historical(self) -> DataFrame:
        return self._df_historical

    @property
    def df_by_measure(self) -> DataFrame:
        return self._df_by_measure

    @property
    def df_by_key(self) -> DataFrame:
        return self._df_by_key



# class PredictabilityBeanLegacy:
#
#     def __init__(
#             self,
#             df_by_measure: DataFrame,
#             df_by_period: DataFrame,
#             df_by_key: DataFrame,
#     ):
#         self._df_by_measure = df_by_measure
#         self._df_by_period = df_by_period
#         self._df_by_key = df_by_key
#
#     @property
#     def df_by_measure(self) -> DataFrame:
#         return self._df_by_measure
#
#     @property
#     def df_by_period(self) -> DataFrame:
#         return self._df_by_period
#
#     @property
#     def df_by_key(self) -> DataFrame:
#         return self._df_by_key



import re
from inspect import stack
from logging import getLogger
from re import Match

from xlsxwriter.utility import xl_cell_to_rowcol, xl_rowcol_to_cell


class CellNotationHelper:
    """
    1. get_address_pattern
    2. is_pattern_xxx
    3. is_address_xxx_pattern
    4. is_address_xxx_correct
    5. transform_xxx_into_xxx
    6. translate_xxx_to_reference

    Old test classes to reshuffle:
    - get_absolute_from_relative_address
    - transform_excel_expression
    """

    # ################################################################################################################ #
    # Fully tested in test_function__get_address_pattern                                                             # #
    # ################################################################################################################ #

    @staticmethod
    def get_address_pattern(str_address: str) -> str:
        """
        Function that analyzes an address, and returns the pattern:
        - cells: --A--1, --A-$1, --A@&1
        - range: --A--1:--A--1
        - column: --A:--A
        - row: --1:--1

        In case the address being analyzed is incorrect, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        str_the_return: str = ""

        # 1. We check the address as a cell
        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*)$",
            string=str_address
        )
        if _match_object:

            if _match_object.group(1) == "":
                str_the_return = str_the_return + "-"
            else:
                str_the_return = str_the_return + "@"

            if _match_object.group(2) == "":
                str_the_return = str_the_return + "-"
            else:
                str_the_return = str_the_return + "$"

            str_the_return = str_the_return + "A"

            if _match_object.group(4) == "":
                str_the_return = str_the_return + "-"
            else:
                str_the_return = str_the_return + "@"

            if _match_object.group(5) == "":
                str_the_return = str_the_return + "-"
            else:
                str_the_return = str_the_return + "$"

            str_the_return = str_the_return + "1"

        else:

            # 2. We check the address as a range
            _match_object: Match = re.match(
                pattern=r"^(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*):(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*)$",
                string=str_address
            )
            if _match_object:
                _lst_tokens: list[str] = str_address.split(":")
                str_the_return = \
                    CellNotationHelper.get_address_pattern(str_address=_lst_tokens[0]) +\
                    ":" +\
                    CellNotationHelper.get_address_pattern(str_address=_lst_tokens[1])

            else:
                # 3. We check the address as a column: --A:--A
                _match_object: Match = re.match(
                    pattern=r"^(\@?)(\$?)([A-Z]+):(\@?)(\$?)([A-Z]+)$",
                    string=str_address
                )
                if _match_object:

                    if _match_object.group(1) == "":
                        str_the_return = str_the_return + "-"
                    else:
                        str_the_return = str_the_return + "@"

                    if _match_object.group(2) == "":
                        str_the_return = str_the_return + "-"
                    else:
                        str_the_return = str_the_return + "$"

                    str_the_return = str_the_return + "A:"

                    if _match_object.group(4) == "":
                        str_the_return = str_the_return + "-"
                    else:
                        str_the_return = str_the_return + "@"

                    if _match_object.group(5) == "":
                        str_the_return = str_the_return + "-"
                    else:
                        str_the_return = str_the_return + "$"

                    str_the_return = str_the_return + "A"

                else:

                    # 4. We check the address as a row: --1:--1
                    _match_object: Match = re.match(
                        pattern=r"^(\@?)(\$?)([1-9][0-9]*):(\@?)(\$?)([1-9][0-9]*)$",
                        string=str_address
                    )
                    if _match_object:

                        if _match_object.group(1) == "":
                            str_the_return = str_the_return + "-"
                        else:
                            str_the_return = str_the_return + "@"

                        if _match_object.group(2) == "":
                            str_the_return = str_the_return + "-"
                        else:
                            str_the_return = str_the_return + "$"

                        str_the_return = str_the_return + "1:"

                        if _match_object.group(4) == "":
                            str_the_return = str_the_return + "-"
                        else:
                            str_the_return = str_the_return + "@"

                        if _match_object.group(5) == "":
                            str_the_return = str_the_return + "-"
                        else:
                            str_the_return = str_the_return + "$"

                        str_the_return = str_the_return + "1"

                    else:
                        raise Exception(f"The address '{str_address}' is not recognized as an address, and its pattern "
                                        f"could not be determined.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return str_the_return

    # ################################################################################################################ #
    # Fully tested in test_function__is_pattern_xxx                                                                  # #
    # ################################################################################################################ #

    @staticmethod
    def is_pattern_correct(str_pattern: str) -> bool:
        """
        Function that returns whether the pattern is correct.

        In case the pattern is incorrect, no exception is raised, but False is returned.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        _match_object: Match = re.match(
            pattern=r"^(\@|-)(\$|-)A(\@|-)(\$|-)1$|"                                # Cell
                    r"^(\@|-)(\$|-)A(\@|-)(\$|-)1:(\@|-)(\$|-)A(\@|-)(\$|-)1$|"     # Range
                    r"^(\@|-)(\$|-)A:(\@|-)(\$|-)A$|"                               # Column
                    r"^(\@|-)(\$|-)1:(\@|-)(\$|-)1$",                               # Row
            string=str_pattern
        )
        if _match_object:
            bool_the_return = True

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_pattern_cell(str_pattern: str) -> bool:
        """
        Function that returns whether the pattern corresponds to a cell: --A--1, --A-$1, --A@&1

        In case the pattern is incorrect, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        if CellNotationHelper.is_pattern_correct(str_pattern=str_pattern):

            _match_object: Match = re.match(
                pattern=r"^(\@|-)(\$|-)A(\@|-)(\$|-)1$",
                string=str_pattern
            )
            if _match_object:
                bool_the_return = True

        else:
            raise Exception(f"Pattern '{str_pattern}' is incorrect, and not a cell!")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_pattern_range(str_pattern: str) -> bool:
        """
        Function that returns whether the pattern corresponds to a range: --A--1:--A--1

        In case the pattern is incorrect, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        if CellNotationHelper.is_pattern_correct(str_pattern=str_pattern):

            _match_object: Match = re.match(
                pattern=r"^(\@|-)(\$|-)A(\@|-)(\$|-)1:(\@|-)(\$|-)A(\@|-)(\$|-)1$",
                string=str_pattern
            )
            if _match_object:
                bool_the_return = True

        else:
            raise Exception(f"Pattern '{str_pattern}' is incorrect, and not a range!")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_pattern_column(str_pattern: str) -> bool:
        """
        Function that returns whether the pattern corresponds to a column: --A:--A

        In case the pattern is incorrect, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        if CellNotationHelper.is_pattern_correct(str_pattern=str_pattern):

            _match_object: Match = re.match(
                pattern=r"^(\@|-)(\$|-)A:(\@|-)(\$|-)A$",
                string=str_pattern
            )
            if _match_object:
                bool_the_return = True

        else:
            raise Exception(f"Pattern '{str_pattern}' is incorrect, and not a column!")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_pattern_row(str_pattern: str) -> bool:
        """
        Function that returns whether the pattern corresponds to a row: --1:--1

        In case the pattern is incorrect, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        if CellNotationHelper.is_pattern_correct(str_pattern=str_pattern):

            _match_object: Match = re.match(
                pattern=r"^(\@|-)(\$|-)1:(\@|-)(\$|-)1$",
                string=str_pattern
            )
            if _match_object:
                bool_the_return = True

        else:
            raise Exception(f"Pattern '{str_pattern}' is incorrect, and not a row!")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    # ################################################################################################################ #
    # Fully tested in test_function__is_address_xxx_pattern                                                          # #
    # ################################################################################################################ #

    @staticmethod
    def is_address_pattern_correct(str_address: str) -> bool:
        """
        Function that returns whether the address pattern is correct.

        In case the address is incorrect, no exception is raised, but False is returned.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = False

        # noinspection PyBroadException
        try:
            CellNotationHelper.get_address_pattern(str_address=str_address)
            bool_the_return = True
        except Exception:
            pass

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_address_cell_pattern(str_address: str) -> bool:
        """
        Function that returns whether the address pattern is a cell.

        In case the address is not a correct one, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = CellNotationHelper.is_pattern_cell(
            str_pattern=CellNotationHelper.get_address_pattern(str_address=str_address)
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_address_range_pattern(str_address: str) -> bool:
        """
        Function that returns whether the address pattern is a range.

        In case the address is not a correct one, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = CellNotationHelper.is_pattern_range(
            str_pattern=CellNotationHelper.get_address_pattern(str_address=str_address)
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_address_column_pattern(str_address: str) -> bool:
        """
        Function that returns whether the address pattern is a column.

        In case the address is not a correct one, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = CellNotationHelper.is_pattern_column(
            str_pattern=CellNotationHelper.get_address_pattern(str_address=str_address)
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    @staticmethod
    def is_address_row_pattern(str_address: str) -> bool:
        """
        Function that returns whether the address pattern is a row.

        In case the address is not a correct one, an exception is raised.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        bool_the_return: bool = CellNotationHelper.is_pattern_row(
            str_pattern=CellNotationHelper.get_address_pattern(str_address=str_address)
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return bool_the_return

    # ################################################################################################################ #
    # Fully tested in test_function__is_address_xxx_correct                                                          # #
    # ################################################################################################################ #

    @staticmethod
    def is_address_correct(str_address: str) -> bool:
        """
        Function that returns whether the address is correct: both patterns and values.

        Exceptions are raised in case:
        - str_address pattern is not correct
        - str_address is dynamic
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _bool_the_return: bool = False

        try:
            _str_the_pattern = CellNotationHelper.get_address_pattern(str_address=str_address)
        except Exception as an_exception:
            raise Exception(f"The parameter str_address '{str_address}' could not be pattern-ized.") from an_exception

        if CellNotationHelper.is_pattern_cell(str_pattern=_str_the_pattern):
            _bool_the_return = CellNotationHelper.is_address_cell_correct(str_address=str_address)
        elif CellNotationHelper.is_pattern_range(str_pattern=_str_the_pattern):
            _bool_the_return = CellNotationHelper.is_address_range_correct(str_address=str_address)
        elif CellNotationHelper.is_pattern_column(str_pattern=_str_the_pattern):
            _bool_the_return = CellNotationHelper.is_address_column_correct(str_address=str_address)
        elif CellNotationHelper.is_pattern_row(str_pattern=_str_the_pattern):
            _bool_the_return = CellNotationHelper.is_address_row_correct(str_address=str_address)
        else:
            raise Exception(f"Could not process the address '{str_address}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _bool_the_return

    @staticmethod
    def is_address_cell_correct(str_address: str) -> bool:
        """
        Function that returns whether the address is a correct cell: no particular business rule

        Exceptions are raised in case:
        - str_address pattern is not correct
        - str_address pattern is not cell
        - str_address is dynamic
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _bool_the_return: bool = False
        _str_the_pattern: str = ""

        # An exception is raised in case the address cannot be pattern-ized
        _str_the_pattern = CellNotationHelper.get_address_pattern(str_address=str_address)

        # In case the pattern is not a cell, we raise an exception
        if not CellNotationHelper.is_pattern_cell(str_pattern=_str_the_pattern):
            raise Exception(f"The parameter str_address '{str_address}' is not a cell.")

        # If the address is dynamic, we can't check if it is correct...
        if "@" in str_address:
            raise Exception(f"The parameter str_address '{str_address}' contains one (or more) '@', and we cannot "
                            f"confirm if it is a correct address or not. Make sure you provide address without any "
                            f"'@'.")

        # No other rule to check for cells...
        _bool_the_return = True

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _bool_the_return

    @staticmethod
    def is_address_range_correct(str_address: str) -> bool:
        """
        Function that returns whether the address is a correct range:
        - Column 1 should be lower or equal to Column 2
        - Row 1 should be lower or equal to Row 2.

        Exceptions are raised in case:
        - str_address pattern is not correct
        - str_address pattern is not range
        - str_address is dynamic
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _bool_the_return: bool = False
        _str_the_pattern: str = ""

        # An exception is raised in case the address cannot be pattern-ized
        _str_the_pattern = CellNotationHelper.get_address_pattern(str_address=str_address)

        # In case the pattern is not a range, we raise an exception
        if not CellNotationHelper.is_pattern_range(str_pattern=_str_the_pattern):
            raise Exception(f"The parameter str_address '{str_address}' is not a range.")

        # If the address is dynamic, we can't check if it is correct...
        if "@" in str_address:
            raise Exception(f"The parameter str_address '{str_address}' contains one (or more) '@', and we cannot "
                            f"confirm if it is a correct address or not. Make sure you provide address without any "
                            f"'@'.")

        # We remove the dollar signs, if any, and tokenize the range into 2 cells
        _lst_str_cells: list[str] = str_address.replace("$", "").split(":")

        # We compare the columns, then the rows
        _str_col_1: str = re.sub("\\d", "", _lst_str_cells[0])
        _str_col_2: str = re.sub("\\d", "", _lst_str_cells[1])

        if len(_str_col_1) < len(_str_col_2):
            _bool_the_return = True
        elif len(_str_col_1) == len(_str_col_2):
            if _str_col_1 <= _str_col_2:
                _bool_the_return = True

        if _bool_the_return:
            # For the rows, we have to compare integers, not strings
            _int_row_1: int = int(re.sub("[A-Z]", "", _lst_str_cells[0]))
            _int_row_2: int = int(re.sub("[A-Z]", "", _lst_str_cells[1]))

            if _int_row_1 <= _int_row_2:
                _bool_the_return = True

            else:
                _bool_the_return = False

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _bool_the_return

    @staticmethod
    def is_address_column_correct(str_address: str) -> bool:
        """
        Function that returns whether the address is a correct column: column 1 should be lower or equal to Column 2.

        Exceptions are raised in case:
        - str_address pattern is not correct
        - str_address pattern is not column
        - str_address is dynamic
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _bool_the_return: bool = False
        _str_the_pattern: str = ""

        # An exception is raised in case the address cannot be pattern-ized
        _str_the_pattern = CellNotationHelper.get_address_pattern(str_address=str_address)

        # In case the pattern is not a column, we raise an exception
        if not CellNotationHelper.is_pattern_column(str_pattern=_str_the_pattern):
            raise Exception(f"The parameter str_address '{str_address}' is not a column.")

        # If the address is dynamic, we can't check if it is correct...
        if "@" in str_address:
            raise Exception(f"The parameter str_address '{str_address}' contains one (or more) '@', and we cannot "
                            f"confirm if it is a correct address or not. Make sure you provide address without any "
                            f"'@'.")

        # We remove the dollar signs, if any, and tokenize the column range into 2 columns
        _lst_str_columns: list[str] = str_address.replace("$", "").split(":")

        # We compare the columns
        _str_col_1: str = re.sub("\\d", "", _lst_str_columns[0])
        _str_col_2: str = re.sub("\\d", "", _lst_str_columns[1])

        if len(_str_col_1) < len(_str_col_2):
            _bool_the_return = True
        elif len(_str_col_1) == len(_str_col_2):
            if _str_col_1 <= _str_col_2:
                _bool_the_return = True

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _bool_the_return

    @staticmethod
    def is_address_row_correct(str_address: str) -> bool:
        """
        Function that returns whether the address is a correct row: row 1 should be lower or equal to row 2.

        Exceptions are raised in case:
        - str_address pattern is not correct
        - str_address pattern is not column
        - str_address is dynamic
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _bool_the_return: bool = False
        _str_the_pattern: str = ""

        # An exception is raised in case the address cannot be pattern-ized
        _str_the_pattern = CellNotationHelper.get_address_pattern(str_address=str_address)

        # In case the pattern is not a row, we raise an exception
        if not CellNotationHelper.is_pattern_row(str_pattern=_str_the_pattern):
            raise Exception(f"The parameter str_address '{str_address}' is not a row.")

        # If the address is dynamic, we can't check if it is correct...
        if "@" in str_address:
            raise Exception(f"The parameter str_address '{str_address}' contains one (or more) '@', and we cannot "
                            f"confirm if it is a correct address or not. Make sure you provide address without any "
                            f"'@'.")

        # We remove the dollar signs, if any, and tokenize the row range into 2 row
        _lst_str_rows: list[str] = str_address.replace("$", "").split(":")

        # For the rows, we have to compare integers, not strings
        _int_row_1: int = int(re.sub("[A-Z]", "", _lst_str_rows[0]))
        _int_row_2: int = int(re.sub("[A-Z]", "", _lst_str_rows[1]))

        if _int_row_1 <= _int_row_2:
            _bool_the_return = True

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _bool_the_return

    # ################################################################################################################ #
    # Fully tested in test_function__transform_xxx_into_xxx                                                          # #
    # ################################################################################################################ #

    @staticmethod
    def transform_address_into_tokens(
            str_address: str,
    ) -> dict:
        """
        Function that transforms a tokenized address into a string address.

        An exception is thrown in case the tokens does not have all the expected keys (either more or less).
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with "
                                  f"str_address = '{str_address}'.")

        _dict_the_return: dict = {
            "first_col": None, "first_col_at": None, "first_col_doll": None,
            "first_row": None, "first_row_at": None, "first_row_doll": None,
            "last_col": None, "last_col_at": None, "last_col_doll": None,
            "last_row": None, "last_row_at": None, "last_row_doll": None,
        }

        # 1. We check the address as a cell
        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*)$",
            string=str_address
        )
        if _match_object:

            _dict_the_return["first_col_at"] = (_match_object.group(1) == "@")
            _dict_the_return["first_col_doll"] = (_match_object.group(2) == "$")
            _dict_the_return["first_col"] = _match_object.group(3)

            _dict_the_return["first_row_at"] = (_match_object.group(4) == "@")
            _dict_the_return["first_row_doll"] = (_match_object.group(5) == "$")
            _dict_the_return["first_row"] = _match_object.group(6)

        else:

            # 2. We check the address as a range
            _match_object: Match = re.match(
                pattern=r"^(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*):(\@?)(\$?)([A-Z]+)(\@?)(\$?)([1-9][0-9]*)$",
                string=str_address
            )
            if _match_object:

                _dict_the_return["first_col_at"] = (_match_object.group(1) == "@")
                _dict_the_return["first_col_doll"] = (_match_object.group(2) == "$")
                _dict_the_return["first_col"] = _match_object.group(3)

                _dict_the_return["first_row_at"] = (_match_object.group(4) == "@")
                _dict_the_return["first_row_doll"] = (_match_object.group(5) == "$")
                _dict_the_return["first_row"] = _match_object.group(6)

                _dict_the_return["last_col_at"] = (_match_object.group(7) == "@")
                _dict_the_return["last_col_doll"] = (_match_object.group(8) == "$")
                _dict_the_return["last_col"] = _match_object.group(9)

                _dict_the_return["last_row_at"] = (_match_object.group(10) == "@")
                _dict_the_return["last_row_doll"] = (_match_object.group(11) == "$")
                _dict_the_return["last_row"] = _match_object.group(12)

            else:
                # 3. We check the address as a column: --A:--A
                _match_object: Match = re.match(
                    pattern=r"^(\@?)(\$?)([A-Z]+):(\@?)(\$?)([A-Z]+)$",
                    string=str_address
                )
                if _match_object:

                    _dict_the_return["first_col_at"] = (_match_object.group(1) == "@")
                    _dict_the_return["first_col_doll"] = (_match_object.group(2) == "$")
                    _dict_the_return["first_col"] = _match_object.group(3)

                    _dict_the_return["last_col_at"] = (_match_object.group(4) == "@")
                    _dict_the_return["last_col_doll"] = (_match_object.group(5) == "$")
                    _dict_the_return["last_col"] = _match_object.group(6)

                else:

                    # 4. We check the address as a row: --1:--1
                    _match_object: Match = re.match(
                        pattern=r"^(\@?)(\$?)([1-9][0-9]*):(\@?)(\$?)([1-9][0-9]*)$",
                        string=str_address
                    )
                    if _match_object:

                        _dict_the_return["first_row_at"] = (_match_object.group(1) == "@")
                        _dict_the_return["first_row_doll"] = (_match_object.group(2) == "$")
                        _dict_the_return["first_row"] = _match_object.group(3)

                        _dict_the_return["last_row_at"] = (_match_object.group(4) == "@")
                        _dict_the_return["last_row_doll"] = (_match_object.group(5) == "$")
                        _dict_the_return["last_row"] = _match_object.group(6)

                    else:
                        raise Exception(f"The address '{str_address}' is not recognized as an address, and it cannot "
                                        f"be tokenized.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_dict_the_return}'.")
        return _dict_the_return

    @staticmethod
    def transform_tokens_into_address(
            dict_tokens: dict,
    ) -> str:
        """
        Function that tokenizes an address. An address could be:
        - A cell
        - A range
        - A set of columns
        - A set of rows.

        An exception is thrown in case the address is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with "
                                  f"str_address = '{dict_tokens}'.")

        if {
            "first_col", "first_col_at", "first_col_doll",
            "first_row", "first_row_at", "first_row_doll",
            "last_col", "last_col_at", "last_col_doll",
            "last_row", "last_row_at", "last_row_doll"
        } != set(dict_tokens.keys()):
            raise Exception(f"Unexpected keys in the parameters dict_tokens.keys()='{', '.join(dict_tokens.keys())}'.")

        _str_the_return: str = ""

        # If we have a first col
        if dict_tokens["first_col"] is not None:
            if dict_tokens["first_col_at"]:
                _str_the_return += "@"
            if dict_tokens["first_col_doll"]:
                _str_the_return += "$"
            _str_the_return += dict_tokens["first_col"]

        # If we have a first row
        if dict_tokens["first_row"] is not None:
            if dict_tokens["first_row_at"]:
                _str_the_return += "@"
            if dict_tokens["first_row_doll"]:
                _str_the_return += "$"
            _str_the_return += dict_tokens["first_row"]

        if (dict_tokens["last_col"] is not None) | (dict_tokens["last_row"] is not None):
            _str_the_return += ":"

        # If we have a last col
        if dict_tokens["last_col"] is not None:
            if dict_tokens["last_col_at"]:
                _str_the_return += "@"
            if dict_tokens["last_col_doll"]:
                _str_the_return += "$"
            _str_the_return += dict_tokens["last_col"]

        # If we have a last row
        if dict_tokens["last_row"] is not None:
            if dict_tokens["last_row_at"]:
                _str_the_return += "@"
            if dict_tokens["last_row_doll"]:
                _str_the_return += "$"
            _str_the_return += dict_tokens["last_row"]

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_str_the_return}'.")
        return _str_the_return

    @staticmethod
    def transform_address_into_relative_and_absolute_address(
            str_address: str,
            b_first_col_at: bool = None, b_first_col_doll: bool = None,
            b_last_col_at: bool = None, b_last_col_doll: bool = None,
            b_first_row_at: bool = None, b_first_row_doll: bool = None,
            b_last_row_at: bool = None, b_last_row_doll: bool = None,
    ):
        """
        Customize an address, to set or unset the @ or $.

        Exceptions are thrown when:
        - The parameter str_address is not correct
        - Whenever we try to set an attribute which does not exist. Eg. b_first_row_at while str_address is a cell...
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _dict_the_tokens: dict = CellNotationHelper.transform_address_into_tokens(str_address=str_address)

        if b_first_col_at is not None:
            if _dict_the_tokens["first_col"] is None:
                raise Exception(f"Trying to set first_col_at='{b_first_col_at}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["first_col_at"] = b_first_col_at

        if b_first_col_doll is not None:
            if _dict_the_tokens["first_col"] is None:
                raise Exception(f"Trying to set first_col_doll='{b_first_col_doll}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["first_col_doll"] = b_first_col_doll

        if b_first_row_at is not None:
            if _dict_the_tokens["first_row"] is None:
                raise Exception(f"Trying to set first_row_at='{b_first_row_at}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["first_row_at"] = b_first_row_at

        if b_first_row_doll is not None:
            if _dict_the_tokens["first_row"] is None:
                raise Exception(f"Trying to set last_row_doll='{b_first_row_doll}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["first_row_doll"] = b_first_row_doll

        if b_last_col_at is not None:
            if _dict_the_tokens["last_col"] is None:
                raise Exception(f"Trying to set last_col_at='{b_last_col_at}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["last_col_at"] = b_last_col_at

        if b_last_col_doll is not None:
            if _dict_the_tokens["last_col"] is None:
                raise Exception(f"Trying to set last_col_doll='{b_last_col_doll}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["last_col_doll"] = b_last_col_doll

        if b_last_row_at is not None:
            if _dict_the_tokens["last_row"] is None:
                raise Exception(f"Trying to set last_row_at='{b_last_row_at}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["last_row_at"] = b_last_row_at

        if b_last_row_doll is not None:
            if _dict_the_tokens["last_row"] is None:
                raise Exception(f"Trying to set last_row_doll='{b_last_row_doll}' for address "
                                f"'{str_address}', impossible.")
            _dict_the_tokens["last_row_doll"] = b_last_row_doll

        _str_the_return: str = CellNotationHelper.transform_tokens_into_address(dict_tokens=_dict_the_tokens)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_str_the_return}'.")
        return _str_the_return

    @staticmethod
    def transform_column_into_range_or_cell(
            str_column: str, int_first_row: int, int_last_row: int,
            b_check_correct: bool = True) -> str:
        """
        Function that returns a range from a column. When the range is one single cell, the function returns the cell.

        Example, for column "C:C" and int_first_row == int_last_row == 3, the function will not return "C3:C3" but "C3".

        Exceptions are raised in case:
        - if b_check_correct, str_column is not a column which is correct (both pattern and value, after having
            removed the "@" if any)
        - if !b_check_correct, str_column is not a column with proper pattern
        - the following condition is not respected: 0 < int_first_row <= int_last_row
        - str_address is dynamic

        TODO: call in this function transform_address_into_relative_and_absolute_address to properly set the row as
            absolute or relative....

        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_tokens: list[str] = str_column.split(":")

        if b_check_correct:
            if not CellNotationHelper.is_address_column_correct(str_address=str_column.replace("@", "")):
                raise Exception(f"The parameter str_address '{str_column}' is not a correct column.")
        else:
            if not CellNotationHelper.is_address_column_pattern(str_address=str_column):
                raise Exception(f"The parameter str_address '{str_column}' is not a column.")

        if (int_first_row <= 0) | (int_first_row > int_last_row):
            raise Exception(f"Rows indices are not correct, and should follow the following rule: "
                            f"0 < int_first_row ('{int_first_row}') <= int_last_row ('{int_last_row}')")

        # We determine whether we need to return a range or a cell
        if (int_first_row == int_last_row) & (_lst_tokens[0] == _lst_tokens[1]):
            str_the_return: str = _lst_tokens[0]+str(int_first_row)
        else:
            str_the_return: str = _lst_tokens[0]+str(int_first_row)+":"+_lst_tokens[1]+str(int_last_row)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return str_the_return

    @staticmethod
    def transform_columns_intersect_rows_into_range_or_cell(str_columns: str, str_rows: str, b_check_correct=True) \
            -> str:
        """
        Function that returns a range from the intersection of a set of columns with a set of rows. When the range is
        one single cell, the function returns the cell.

        Example, intersection of "C:C" and "3:3" will not return "C3:C3" but "C3".

        Exceptions are raised in case:
        - if b_check_correct:
            - str_columns is not a set of columns which is correct (both pattern and value)
            - str_rows is not a set of row which is correct (both pattern and value)
            - returned range is not a range which is correct (both pattern and value)
        - if !b_check_correct:
            - str_column is not a set of columns with proper pattern
            - str_row is not a set of rows with proper pattern
            - returned range is not a range which with proper pattern
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # Before starting, we check the columns and rows are correct
        if b_check_correct:
            if not CellNotationHelper.is_address_column_correct(str_address=str_columns):
                raise Exception(f"The parameter str_columns '{str_columns}' is not a correct set of columns.")
            if not CellNotationHelper.is_address_row_correct(str_address=str_rows):
                raise Exception(f"The parameter str_rows '{str_rows}' is not a correct set of rows.")
        else:
            if not CellNotationHelper.is_address_column_pattern(str_address=str_columns):
                raise Exception(f"The parameter str_columns '{str_columns}' does not have column pattern.")
            if not CellNotationHelper.is_address_row_pattern(str_address=str_rows):
                raise Exception(f"The parameter str_rows '{str_rows}' does not have row pattern.")

        _lst_str_columns: list[str] = str_columns.split(":")
        _lst_str_rows: list[str] = str_rows.split(":")

        _str_the_first_cell: str = _lst_str_columns[0]+_lst_str_rows[0]
        _str_the_second_cell: str = _lst_str_columns[1]+_lst_str_rows[1]

        # Particular case: we do not return A3:A3, but A3
        if _str_the_first_cell == _str_the_second_cell:
            _str_the_return: str = _str_the_first_cell

            if b_check_correct:
                if not CellNotationHelper.is_address_cell_correct(str_address=_str_the_return):
                    raise Exception(f"The returned value '{_str_the_return}' is not a correct cell.")
            else:
                if not CellNotationHelper.is_address_cell_pattern(str_address=_str_the_return):
                    raise Exception(f"The returned value '{_str_the_return}' does not have cell pattern.")

        # Else, both cells are distinct, we return a range...
        else:
            _str_the_return: str = _str_the_first_cell + ":" + _str_the_second_cell

            if b_check_correct:
                if not CellNotationHelper.is_address_range_correct(str_address=_str_the_return):
                    raise Exception(f"The returned value '{_str_the_return}' is not a correct range.")
            else:
                if not CellNotationHelper.is_address_range_pattern(str_address=_str_the_return):
                    raise Exception(f"The returned value '{_str_the_return}' does not have range pattern.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    # ################################################################################################################ #
    # Fully tested in test_function__translate_xxx_to_reference                                                      # #
    # ################################################################################################################ #

    @staticmethod
    def translate_address_to_reference(str_address: str, str_reference_cell: str) -> str:
        """
        Function that translates an address to a given reference. Only relative components prefixed with "@" are
        translated.

        Exceptions are raised in case:
        - str_address is not an address
        - str_reference_cell is not a cell which is correct
        - the resulting translated range is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _str_the_return: str

        if CellNotationHelper.is_address_cell_pattern(str_address=str_address):
            _str_the_return = CellNotationHelper.translate_cell_to_reference(
                str_cell=str_address,
                str_reference_cell=str_reference_cell
            )

        elif CellNotationHelper.is_address_range_pattern(str_address=str_address):
            _str_the_return = CellNotationHelper.translate_range_to_reference(
                str_range=str_address,
                str_reference_cell=str_reference_cell
            )

        elif CellNotationHelper.is_address_column_pattern(str_address=str_address):
            _str_the_return = CellNotationHelper.translate_column_to_reference(
                str_column=str_address,
                str_reference_cell=str_reference_cell
            )

        elif CellNotationHelper.is_address_row_pattern(str_address=str_address):
            _str_the_return = CellNotationHelper.translate_row_to_reference(
                str_row=str_address,
                str_reference_cell=str_reference_cell
            )

        else:
            raise Exception(f"Unexpected case, with str_address='{str_address}' and "
                            f"str_reference_cell='{str_reference_cell}'.")

        if not CellNotationHelper.is_address_correct(str_address=_str_the_return):
            raise Exception(f"The translated address '{_str_the_return}' does not appear to be correct, "
                            f"with str_address='{str_address}' and str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_range_to_reference(str_range: str, str_reference_cell: str) -> str:
        """
        Function that translates a range to a given reference. Only relative components prefixed with "@" are
        translated.

        Exceptions are raised in case:
        - str_range is not a range with proper pattern
        - str_reference_cell is not a cell which is correct
        - the resulting translated range is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        if not CellNotationHelper.is_address_range_pattern(str_address=str_range):
            raise Exception(f"The parameter str_range '{str_range}' has a pattern which is not range.")

        _lst_str_tokens: list[str] = str_range.split(":")

        _str_the_return: str = CellNotationHelper.translate_cell_to_reference(
            str_cell=_lst_str_tokens[0],
            str_reference_cell=str_reference_cell
        ) + ":" + CellNotationHelper.translate_cell_to_reference(
            str_cell=_lst_str_tokens[1],
            str_reference_cell=str_reference_cell
        )

        if not CellNotationHelper.is_address_range_correct(str_address=_str_the_return):
            raise Exception(f"The translated range '{_str_the_return}' does not appear to be correct, "
                            f"with str_range='{str_range}' and str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_column_to_reference(str_column: str, str_reference_cell: str) -> str:
        """
        Function that translates a column range to a given reference. Only relative components prefixed with "@" are
        translated.

        A column range looks like "A:A", and is of the pattern: ??A---:??A---

        Exceptions are raised in case:
        - str_column is not a column range with proper pattern
        - str_reference_cell is not a cell which is correct
        - the resulting translated column range is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        if not CellNotationHelper.is_address_column_pattern(str_address=str_column):
            raise Exception(f"The parameter str_column '{str_column}' has a pattern which is not column range.")

        _lst_str_tokens: list[str] = str_column.split(":")

        _str_the_return: str = CellNotationHelper.translate_column_ref_to_reference(
            str_column_ref=_lst_str_tokens[0],
            str_reference_cell=str_reference_cell
        ) + ":" + CellNotationHelper.translate_column_ref_to_reference(
            str_column_ref=_lst_str_tokens[1],
            str_reference_cell=str_reference_cell
        )

        if not CellNotationHelper.is_address_column_correct(str_address=_str_the_return):
            raise Exception(f"The translated range '{_str_the_return}' does not appear to be correct, "
                            f"with str_column='{str_column}' and str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_row_to_reference(str_row: str, str_reference_cell: str) -> str:
        """
        Function that translates a row range to a given reference. Only relative components prefixed with "@" are
        translated.

        A row range looks like "1:1", and is of the pattern: ---??1:---??1

        Exceptions are raised in case:
        - str_row is not a row range with proper pattern
        - str_reference_cell is not a cell which is correct
        - the resulting translated row range is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        if not CellNotationHelper.is_address_row_pattern(str_address=str_row):
            raise Exception(f"The parameter str_row '{str_row}' has a pattern which is not row range.")

        _lst_str_tokens: list[str] = str_row.split(":")

        _str_the_return: str = CellNotationHelper.translate_row_ref_to_reference(
            str_row_ref=_lst_str_tokens[0],
            str_reference_cell=str_reference_cell
        ) + ":" + CellNotationHelper.translate_row_ref_to_reference(
            str_row_ref=_lst_str_tokens[1],
            str_reference_cell=str_reference_cell
        )

        if not CellNotationHelper.is_address_row_correct(str_address=_str_the_return):
            raise Exception(f"The translated range '{_str_the_return}' does not appear to be correct, "
                            f"with str_row='{str_row}' and str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_cell_to_reference(str_cell: str, str_reference_cell: str) -> str:
        """
        Function that translates a cell to a given reference. Only relative components prefixed with "@" are
        translated.

        Exceptions are raised in case:
        - str_cell does not have a correct pattern
        - str_reference_cell is not a cell which is correct
        - the resulting translated range is not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        if not CellNotationHelper.is_address_cell_pattern(str_address=str_cell):
            raise Exception(f"The parameter str_cell '{str_cell}' does not have a cell pattern.")

        _str_the_return: str = str_cell

        # If the address is dynamic
        if "@" in str_cell:

            # We split the cell into 2, the column and the row
            _match_object: Match = re.match(
                pattern=r"^(\@?\$?[A-Z]+)(\@?\$?[1-9][0-9]*)$",
                string=str_cell
            )

            _str_the_return = CellNotationHelper.translate_column_ref_to_reference(
                str_column_ref=_match_object.group(1),
                str_reference_cell=str_reference_cell
            )+CellNotationHelper.translate_row_ref_to_reference(
                str_row_ref=_match_object.group(2),
                str_reference_cell=str_reference_cell
            )

        if not CellNotationHelper.is_address_cell_correct(str_address=_str_the_return):
            raise Exception(f"The translated range '{_str_the_return}' does not appear to be correct, "
                            f"with str_cell='{str_cell}' and str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_column_ref_to_reference(str_column_ref: str, str_reference_cell: str) -> str:
        """
        Function that translates a column ref to a given reference, only if prefixed with "@".

        A column ref is "C", but not "C:C".

        Exceptions are raised in case:
        - str_column_ref is not a correct alpha (possibly with $)
        - str_reference_cell is not a cell which is correct
        - the resulting column ref is not a correct alpha (possibly with $)
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _str_the_return: str = str_column_ref

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([A-Z]+)$",
            string=str_column_ref
        )
        if not _match_object:
            raise Exception(f"The parameter str_column_ref '{str_column_ref}' is not correct.")

        # In case the column reference is dynamic
        # E.g. str_column_ref="@$D" and str_reference_cell="B3"
        if "@" in str_column_ref:

            # We get "D"
            _str_the_pure_column_ref: str = re.sub(r"[@$]", "", str_column_ref)

            # We get "D1"
            _str_the_column_ref_as_cell: str = _str_the_pure_column_ref+"1"

            # We get "4" from "(1, 4)"
            _int_the_column_ref_as_col_index: int = xl_cell_to_rowcol(cell_str=_str_the_column_ref_as_cell)[1]

            # We get "2" from "(3, 2)"
            _int_the_reference_cell_as_col_index: int = xl_cell_to_rowcol(cell_str=str_reference_cell)[1]

            # We get "F1"
            _str_the_translated_cell = xl_rowcol_to_cell(
                row=1,
                col=_int_the_column_ref_as_col_index+_int_the_reference_cell_as_col_index
            )

            # We get "F"
            _str_the_translated_column_ref = re.sub("[0-9]", "", _str_the_translated_cell)

            # We get $F
            if "$" in str_column_ref:
                _str_the_return = "$"+_str_the_translated_column_ref
            else:
                _str_the_return = _str_the_translated_column_ref

        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([A-Z]+)$",
            string=_str_the_return
        )
        if not _match_object:
            raise Exception(f"The resulting column_ref '{_str_the_return}' is not correct. It was computed from "
                            f"the parameters str_column_ref='{str_column_ref}' and "
                            f"str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    @staticmethod
    def translate_row_ref_to_reference(str_row_ref: str, str_reference_cell: str) -> str:
        """
        Function that translates a row reference to a given reference, only if prefixed with "@".

        A row ref is "1", but not "1:1".

        Exceptions are raised in case:
        - str_row_ref is not a correct numeric (possibly with $)
        - str_reference_cell is not a cell which is correct
        - the resulting row ref is not a correct numeric (possibly with $)
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _str_the_return: str = str_row_ref

        if not CellNotationHelper.is_address_cell_correct(str_address=str_reference_cell):
            raise Exception(f"The parameter str_reference '{str_reference_cell}' is not correct.")

        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([1-9][0-9]*)$",
            string=str_row_ref
        )
        if not _match_object:
            raise Exception(f"The parameter str_row_ref '{str_row_ref}' is not correct.")

        # In case the row reference is dynamic
        # E.g. str_row_ref="@$3" and str_reference_cell="B3"
        if "@" in str_row_ref:

            # We get "3" as a numeric
            _int_the_row_ref: int = int(re.sub(r"[@$]", "", str_row_ref))

            # We get "3" from "(3, 2)"
            _int_the_reference_cell_as_row_index: int = xl_cell_to_rowcol(cell_str=str_reference_cell)[0]

            # We get $F
            if "$" in str_row_ref:
                _str_the_return = "$"+str(_int_the_reference_cell_as_row_index+_int_the_row_ref)
            else:
                _str_the_return = str(_int_the_reference_cell_as_row_index+_int_the_row_ref)

        _match_object: Match = re.match(
            pattern=r"^(\@?)(\$?)([1-9][0-9]*)$",
            string=_str_the_return
        )
        if not _match_object:
            raise Exception(f"The resulting row_ref '{_str_the_return}' is not correct. It was computed from "
                            f"the parameters str_row_ref='{str_row_ref}' and "
                            f"str_reference_cell='{str_reference_cell}'.")

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _str_the_return

    # ################################################################################################################ #
    # Not tested yet                                                                  # #
    # ################################################################################################################ #

    @staticmethod
    def translate_formula_to_reference(str_reference_address: str, str_formula: str) -> str:
        """
        Function that translates a formula to a given reference, only if it contains addresses prefixed with "@".

        Exceptions are raised in case:
        - str_formula cannot be successfully parsed
        - translated references are not correct
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        str_the_return: str = ""

        # We tokenize the expression, to isolate the "string" tokens to ignore
        _lst_the_expression_tokens: list[str] = CellNotationHelper._translate_formula_to_reference__tokenize(
            str_formula=str_formula
        )

        # For each token
        for str_i_token in _lst_the_expression_tokens:

            # if it starts with ["], it means the token is a string parameter, not to be processed
            if str_i_token.startswith('"'):
                str_the_return = str_the_return + str_i_token

            # else, the token is not a string
            else:

                # 1. We process cells references of the kind "A3"
                str_i_transformed_token: str = CellNotationHelper._translate_formula_to_reference__pattern_a1(
                    str_reference_address=str_reference_address,
                    str_formula_token=str_i_token
                )

                # 2. We process cells references of the kind "A:A"
                str_i_transformed_token = CellNotationHelper._translate_formula_to_reference__pattern_aa(
                    str_reference_address=str_reference_address,
                    str_formula_token=str_i_transformed_token
                )

                # 3. We process cells references of the kind "1:1"
                str_i_transformed_token = CellNotationHelper._translate_formula_to_reference__pattern_11(
                    str_reference_address=str_reference_address,
                    str_formula_token=str_i_transformed_token
                )

                # 4. Eventually, we concatenate
                str_the_return = str_the_return + str_i_transformed_token

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return str_the_return

    @staticmethod
    def _translate_formula_to_reference__tokenize(str_formula: str) -> list[str]:
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        _lst_the_return: list[str] = re.split(
            pattern=r'("(?:""|[^"])*")',
            string=str_formula
        )

        # We check the concatenation of the tokens is effectively equal to the
        # initial expression
        if str_formula != "". join(_lst_the_return):
            raise Exception(
                f"When calling function re.split(r'(\"(?:(?:\"\")|(?:[^\"]))*\")', '{str_formula}'), we obtained "
                f"a list of tokens '{', '.join(_lst_the_return)}', concatenating in a string "
                f"'{''. join(_lst_the_return)}' which is not equal to the initial expression."
            )
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return _lst_the_return

    @staticmethod
    def _translate_formula_to_reference__pattern_a1(str_reference_address: str, str_formula_token: str) -> str:
        """
        The regular expression is long and can be split as below:
        - (?<!\\[)\\@: matches special character "@" if not preceded by "[": this notation can be used in formulas in
          tables, and should not be altered
        - The first half of the regex, up to the "|", will match reference such as: @A1
        - The second half of the regex, from the "|", will match reference such as: A@1
        - we use "[1-9][0-9]*" rather then "\\d+", as we cannot have a cell reference which is A0

        :param str_reference_address:
        :param str_formula_token:
        :return:
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_str_cell_references: list[str] = re.findall(
            pattern=r"(?<!\[)\@\$?[A-Z]+\@?\$?[1-9][0-9]*|(?<!\[)\@?\$?[A-Z]+\@\$?[1-9][0-9]*",
            string=str_formula_token
        )

        str_the_return: str = str_formula_token

        for str_i_cell_ref in _lst_str_cell_references:
            _str_transformed_cell_ref: str = CellNotationHelper.translate_cell_to_reference(
                str_reference_cell=str_reference_address,
                str_cell=str_i_cell_ref
            )
            str_the_return = str_the_return.replace(str_i_cell_ref, _str_transformed_cell_ref, 1)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return str_the_return

    @staticmethod
    def _translate_formula_to_reference__pattern_aa(str_reference_address: str, str_formula_token: str) -> str:
        """
        The regular expression is long and can be split as below:
        - (?<!|\\[)|\\@: matches special character "@" if not preceded by "[": this notation can be used in formulas in
          tables, and should not be altered
        - The first half of the regex, up to the "|", will match reference such as: @A:A
        - The second half of the regex, from the "|", will match reference such as: A:@A

        :param str_reference_address:
        :param str_formula_token:
        :return:
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_str_cell_references: list[str] = re.findall(
            pattern=r"(?<!\[)\@\$?[A-Z]+:\@?\$?[A-Z]+|(?<!\[)\@?\$?[A-Z]+:\@\$?[A-Z]+",
            string=str_formula_token
        )

        str_the_return: str = str_formula_token

        for str_i_column_ref in _lst_str_cell_references:
            _str_transformed_column_ref: str = CellNotationHelper.translate_column_to_reference(
                str_reference_cell=str_reference_address,
                str_column=str_i_column_ref
            )
            str_the_return = str_the_return.replace(str_i_column_ref, _str_transformed_column_ref, 1)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return str_the_return

    @staticmethod
    def _translate_formula_to_reference__pattern_11(str_reference_address: str, str_formula_token: str) -> str:
        """
        The regular expression is long and can be split as below:
        - (?<!\\[)\\@: matches special character "@" if not preceded by "[": this notation can be used in formulas in
          tables, and should not be altered
        - The first half of the regex, up to the "|", will match reference such as: @1:1
        - The second half of the regex, from the "|", will match reference such as: 1:@1

        :param str_reference_address:
        :param str_formula_token:
        :return:
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_str_cell_references: list[str] = re.findall(
            pattern=r"(?<!\[)\@\$?[1-9][0-9]*:\@?\$?[1-9][0-9]*|(?<!\[)\@?\$?[1-9][0-9]*:\@\$?[1-9][0-9]*",
            # pattern=r"\@\$?[1-9][0-9]*:\@?\$?[1-9][0-9]*|\@?\$?[1-9][0-9]*:\@\$?[1-9][0-9]*",
            string=str_formula_token
        )

        str_the_return: str = str_formula_token

        for str_i_row_ref in _lst_str_cell_references:
            _str_transformed_row_ref: str = CellNotationHelper.translate_row_to_reference(
                str_reference_cell=str_reference_address,
                str_row=str_i_row_ref
            )
            str_the_return = str_the_return.replace(str_i_row_ref, _str_transformed_row_ref, 1)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return str_the_return

    # @staticmethod
    # def get_absolute_from_relative_address(str_reference_address: str, str_relative_address: str) -> str:
    #     """
    #     Function that translate an address or a range, relatively to a set reference. Only the row/ column prefixed
    #     with an @ is translated.
    #
    #     Illustration:
    #     - Relative to a reference of "B3"
    #     - A relative address of "@C@5" is translated to "D7"
    #     - A relative range of "@$C@$5:D8" is translated to "$D$7:D8"
    #     - A relative range of "@C" or "@C:@D" is translated to "D" or "D:E
    #     - A relative range of "@5" or "@5:@6" is translated to "7" or "7:8"
    #
    #     :param str_reference_address: the address against which the relative addressed is expressed
    #     :param str_relative_address: the address relative to the reference
    #     :return: the address relative to the sheet
    #     """
    #     getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
    #     _str_the_return: str = ""
    #
    #     # If the relative address is a a range, we have to process tokens one by one
    #     if ":" in str_relative_address:
    #
    #         # We recursively call the function on each token
    #         _lst_cells: list[str] = str_relative_address.split(":")
    #         _str_cell_0: str = CellNotationHelper.get_absolute_from_relative_address(
    #             str_reference_address=str_reference_address,
    #             str_relative_address=_lst_cells[0]
    #         )
    #         _str_cell_1: str = CellNotationHelper.get_absolute_from_relative_address(
    #             str_reference_address=str_reference_address,
    #             str_relative_address=_lst_cells[1]
    #         )
    #         _str_the_return = _str_cell_0 + ":" + _str_cell_1
    #
    #     # Else, the relative address is a cell, an entire column or an entire row
    #     else:
    #
    #         # We prepare the return value, removing the special character "@"
    #         _str_the_return = str_relative_address.replace("@", "")
    #
    #         # If the address contains an alpha column
    #         _match_the_column: Match = re.search(r"^@?\$?[A-Z]+", str_relative_address)
    #         if _match_the_column is not None:
    #
    #             # We get the alpha column from the address, possibly with @ and $
    #             _str_the_column: str = _match_the_column.group(0)
    #
    #             # If this alpha column is relative, aka contains an @
    #             if "@" in _str_the_column:
    #
    #                 # We get the alpha column only, excluding all other characters
    #                 _str_the_alpha: str = re.search(r"[A-Z]+$", _str_the_column).group(0)
    #
    #                 _int_the_relative_col: int = xl_cell_to_rowcol(_str_the_alpha + "1")[1]
    #                 _int_the_reference_col: int = xl_cell_to_rowcol(str_reference_address)[1]
    #
    #                 _str_the_new_alpha: str = xl_col_to_name(_int_the_relative_col + _int_the_reference_col)
    #                 _str_the_return: str = _str_the_return.replace(_str_the_alpha, _str_the_new_alpha)
    #
    #         # If the address contains an numeric row
    #         _match_the_row: Match = re.search(r"@?\$?[0-9]+$", str_relative_address)
    #         if _match_the_row is not None:
    #
    #             # We get the numeric row from the address, possibly with @ and $
    #             _str_the_row: str = _match_the_row.group(0)
    #
    #             # If this numeric row is relative, aka contains an @
    #             if "@" in _str_the_row:
    #
    #                 # We get the numeric row only, excluding all other characters
    #                 _int_the_row: int = int(re.search(r"[0-9]+$", _str_the_row).group(0))
    #
    #                 _int_the_reference_row: int = xl_cell_to_rowcol(str_reference_address)[0]
    #
    #                 _int_the_new_numerics: int = _int_the_row + _int_the_reference_row
    #                 _str_the_return: str = _str_the_return.replace(str(_int_the_row), str(_int_the_new_numerics))
    #
    #     getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
    #     return _str_the_return



import json

import click
from click import BadParameter
from pandas import DataFrame

from com.enovation.helper.json_encoder import JSONEncoder

import logging
from inspect import stack


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('df-to-stdout')
@click.pass_context
@click.argument('alias-in', type=str)
def df_to_stdout(ctx_context, alias_in):
    """
    Print the predictabilities.

    :param ctx_context: the context
    :param alias_in: alias to a dataframe to print.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot print '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively loaded some data, and labelled these with the alias '{alias_in}'.",
            ctx=ctx_context
        )

    str_to_print: str = ""
    try:
        str_to_print = ctx_context.obj[alias_in].to_string()
    except Exception as exception:
        _logger.debug(f"ctx_context.obj[{alias_in}].to_string() raised an exception: "+str(exception))
        try:
            str_to_print = json.dumps(
                obj=ctx_context.obj[alias_in],
                skipkeys=True,
                cls=JSONEncoder
            )
        except BaseException as exception:
            _logger.error(f"[DBG] json.dumps(obj=ctx_context.obj[{alias_in}], skipkeys=True, cls=JSONEncoder) raised "
                          f"an exception: " + str(exception))

    if len(str_to_print) > 0:
        click.echo_via_pager(text_or_generator=str_to_print)
    else:
        click.echo("Nothing to print")

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('df-to-xlsx')
@click.pass_context
@click.argument('alias', type=str)
@click.argument('file', type=click.Path(exists=False))
def df_to_xlsx(ctx_context, alias, file):
    """
    Print the predictabilities.

    :param ctx_context: the context.
    :param alias: alias to a dataframe to persist.
    :param file: the file to persist.
    """
    _logger.info(f"\n"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\nCommand df-to-xlsx called with parameters:"
                 f"\n\t- file: {file}"
                 f"\n\t- alias: {alias}")

    if alias not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively loaded some data, and labelled these with the alias '{alias}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias], DataFrame):
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias is of type '{type(ctx_context.obj[alias])}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively used operands to generate data as a DataFrame.",
            ctx=ctx_context
        )

    df_the_data: DataFrame = ctx_context.obj[alias]

    df_the_data.to_excel(excel_writer=file)

    _logger.info(f"\nCommand df-to-xlsx is returning"
                 f"\n**************************************************************************************************"
                 f"\n**************************************************************************************************"
                 f"\n"
                 )



from inspect import stack
from logging import Logger, getLogger

from dash import html

from com.enovation.toolbox.dash.layout.debug.debug__as_div import DebugAsDiv
from com.enovation.toolbox.dash.layout.footer.footer__as_footer import FooterAsFooter
from com.enovation.toolbox.dash.layout.navigation.navigation__as_navbar import NavigationAsNavbar
from com.enovation.toolbox.dash.layout.welcome.welcome__as_div import WelcomeAsDiv
from com.enovation.toolbox.dash.session import Session
from com.enovation.toolbox.dash.sheet import Sheet
import dash_bootstrap_components as dbc

from com.enovation.toolbox.dash.widget import WidgetAsDiv


class LayoutAsDiv(WidgetAsDiv):
    _logger: Logger = getLogger(__name__)

    @property
    def dict_sheets(self) -> dict[str, Sheet]:
        return self._dict_sheets

    # The widget IDs to be used
    # Naming convention:
    # - should start by "const_str__wid__"
    # - as the application will screen all the widget ids, across widgets, to ensure there is no ambiguity, aka same
    #   widget id across different widgets
    const_str__wid__main_content: str = "content"

    def __init__(
            self,
            dict_sheets: dict[str, Sheet],
            dict_config: dict,
            dict_context: dict,
            session: Session
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        super().__init__(
            dict_config=dict_config,
            dict_context=dict_context,
            session=session,
            children=[

                # The navigation bar at the top of the screen
                NavigationAsNavbar(
                    dict_sheet=dict_sheets,
                    str__wid__main_content=self.const_str__wid__main_content,
                    dict_config=dict_config,
                    dict_context=dict_context,
                    session=session
                ),

                # The main page in a container
                dbc.Container(
                    children=[
                        html.Main(
                            id=self.const_str__wid__main_content,
                            className="container",
                            children=WelcomeAsDiv(
                                dict_config=dict_config,
                                dict_context=dict_context,
                                session=session
                            ),
                        ),

                        DebugAsDiv(
                            dict_config=dict_config,
                            dict_context=dict_context,
                            session=session,
                        ),
                    ],
                ),

                # The footer
                FooterAsFooter(
                    dict_config=dict_config,
                    dict_context=dict_context,
                    session=session
                )
            ])

        self._dict_sheets = dict_sheets

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import logging
from datetime import datetime
import json

from pandas import DataFrame, to_datetime, NaT, isnull


class JSONEncoder(json.JSONEncoder):

    _logger: logging.Logger = logging.getLogger(__name__)

    def default(self, obj):

        if isinstance(obj, DataFrame):
            return obj.to_dict()

        elif isinstance(obj, datetime):
            return obj.isoformat()

        # Let the base class default method raise the TypeError
        return JSONEncoder.default(self, obj)



import click
import ast


class PythonLiteralOption(click.Option):

    def type_cast_value(self, ctx, value):

        if isinstance(value, str):
            try:
                return ast.literal_eval(value)
            except Exception as exception:
                raise click.BadParameter(value) from exception
        else:
            # This happen when the default value is already a typed instance (aka not a string)
            return value


class PythonLiteralArgument(click.Argument):

    def type_cast_value(self, ctx, value):

        if isinstance(value, str):
            try:
                return ast.literal_eval(value)
            except Exception as exception:
                raise click.BadParameter(value) from exception
        else:
            # This happen when the default value is already a typed instance (aka not a string)
            return value



import logging
from inspect import stack

import click

from com.enovation.helper.click.df_load import df_load_csv, df_load_xls
from com.enovation.helper.click.df_write import df_to_stdout, df_to_xlsx
from com.enovation.helper.click.df_modify import df_rename_columns, df_cleanse_null, df_compress
from com.enovation.helper.click.dict_load_modify_write import dict_load_json, dict_set_json, dict_modif
from com.enovation.helper.click.py_exec import py_exec_file
from com.enovation.toolbox.dash.click.ds_dash import ds_dash
from com.enovation.toolbox.excel_dashboard.click.dashboard import dashboard
from com.enovation.toolbox.predictability.click.dp_compute import dp_compute_date_predictability
from com.enovation.toolbox.predictability.click.dp_demo import dp_demo_date_predictability
from com.enovation.toolbox.predictability.click.dp_persist \
    import dp_write_bean_date_predictability, dp_load_bean_date_predictability
from com.enovation.toolbox.predictability.click.dp_graph import dp_graph_to_dash_date_predictability

_logger: logging.Logger = logging.getLogger(__name__)


@click.group(chain=True)
@click.pass_context
@click.option(
    '--verbose/--no-verbose', type=bool, default=None,
    help='Level of logging verbosity: INFO (--verbose), WARNING (default) or ERROR (--no-verbose).',
)
def enov(ctx_context, verbose):

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if verbose is True:
        click.echo("logging: INFO")
        logging.basicConfig(level="INFO")
    elif verbose is False:
        click.echo("logging: ERROR")
        logging.basicConfig(level="ERROR")
    else:
        click.echo("logging: WARNING")
        logging.basicConfig(level="WARNING")
    _logger.info(f"Welcome!")

    # Ensure that ctx_context.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    # This is effectively the context, that is shared across commands
    if ctx_context.obj:
        pass
    else:
        _logger.debug(f"We call function ctx_context.ensure_object(dict)")
        ctx_context.ensure_object(dict)

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


# com.enovation.helper.click.df_load
# noinspection PyTypeChecker
enov.add_command(df_load_csv)
# noinspection PyTypeChecker
enov.add_command(df_load_xls)

# com.enovation.helper.click.df_write
# noinspection PyTypeChecker
enov.add_command(df_to_stdout)
# noinspection PyTypeChecker
enov.add_command(df_to_xlsx)

# com.enovation.helper.click.df_modify
# noinspection PyTypeChecker
enov.add_command(df_rename_columns)
# noinspection PyTypeChecker
enov.add_command(df_cleanse_null)
# noinspection PyTypeChecker
enov.add_command(df_compress)

# com.enovation.helper.click.dict_load_modify_write
# noinspection PyTypeChecker
enov.add_command(dict_load_json)
# noinspection PyTypeChecker
enov.add_command(dict_set_json)
# noinspection PyTypeChecker
enov.add_command(dict_modif)

# com.enovation.toolbox.predictability.click - Date Predictability
# noinspection PyTypeChecker
enov.add_command(dp_compute_date_predictability)
# noinspection PyTypeChecker
enov.add_command(dp_write_bean_date_predictability)
# noinspection PyTypeChecker
enov.add_command(dp_load_bean_date_predictability)
# noinspection PyTypeChecker
enov.add_command(dp_graph_to_dash_date_predictability)
# noinspection PyTypeChecker
enov.add_command(dp_demo_date_predictability)

# com.enovation.toolbox.excel_dashboard.click - Dashboarding capabilities
# noinspection PyTypeChecker
enov.add_command(dashboard)

# noinspection PyTypeChecker
enov.add_command(ds_dash)

# noinspection PyTypeChecker
enov.add_command(py_exec_file)


if __name__ == '__main__':
    enov()



from inspect import stack
from logging import Logger, getLogger
from dash import html
from dash.html import Div


class TabWelcomeAsDiv(Div):
    """
    Return the welcome widget such as:
    <div class="jumbotron">
        <h1>Date Predictability</h1>
        <p class="lead">
            'Predictability' is about accurately predicting early enough the project outcomes (incl. beyond own
            roles and responsibilities), to enable timely corrective actions that increase the likelihood of
            achieving targets and reducing outcome variance.
        </p>
        <hr class="my-2">
        <p>Explanation of how it works here...</p>
    </div>
    """
    _logger: Logger = getLogger(__name__)

    def __init__(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        super().__init__(
            children=[
                html.H1("Date Predictability"),
                html.P(
                    "'Predictability' is about accurately predicting early enough the project outcomes "
                    "(incl. beyond own roles and responsibilities), to enable timely corrective actions "
                    "that increase the likelihood of achieving targets and reducing outcome variance.",
                    className="lead",
                ),
                html.Hr(className="my-2"),
                html.P(
                    "Explanation of how it works here..."
                )
            ],
            className="jumbotron",
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack
from logging import Logger, getLogger
from typing import Optional

from xlsxwriter.utility import xl_cell_to_rowcol
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from com.enovation.toolbox.excel_dashboard.widgets.widget import Widget


class WidgetCoverSheet(Widget):
    """
    Simple widget that print a cover sheet in the Excel Spreadsheet.

    Configuration:
    - "narrative": the main text to print in the cover sheet. Texts to infer is handled through {}.
                   Illustration: 'Height: {0:f} {1:s}'.format(172.3, 'cm')

    Parameters:
    - "lst_tokens": list of tokens to infer into the narrative.
    """

    # The labels to parse the json configuration file
    const_str__config_label__str_narrative: str = "narrative"

    # The labels for the parameters expected by the widget
    const_str__parameter_label__lst_tokens: list = "tokens"

    _logger: Logger = getLogger(__name__)

    @property
    def str_narrative(self) -> str:
        return self._str_narrative

    @property
    def lst_tokens(self) -> list:
        return self._lst_tokens

    def __init__(
            self,
            str_address: str,
            dict_parameters_map: dict,
            dict_config: dict,
            dict_raw_parameters: dict,
            str_sheet_name: str,
            dict_default_format: dict
    ):
        """
        Initialize the widget:
        - Calling the Widget super init function
        - Declare the widget properties:
          - Each has their "property" getters
          - They are declared and instantiated to None
          - They are effectively set through the function "check_parameters_and_set_properties"

        :param str_address: the address in a sheet to print the widget
        :param dict_config: the widget configuration (aka the "config" node in the json)
        :param dict_parameters_map: the dictionary to map users parameters to the ones expected by the widget:
                                    - key: the expected label
                                    - value: the actual label
        :param dict_raw_parameters: the parameters as labelled and provided by the users
        :param str_sheet_name: the name of the sheet in which the widget is instantiated (required to handle default
            formats inheritance, from the workbook to the workshoeet to the widget)
        :param dict_default_format: the default format set at the level of the widget. If such default format is not
            configured, we expect an empty dictionary
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We declare the widget properties, initialized to None
        self._str_narrative: Optional[str] = None
        self._lst_tokens: Optional[list] = None

        # The super init function initializes the widget:
        # - Step 1, we set the widget properties (from nodes from the json configuration): address, parameters map,
        #   config
        # - Step 2, we process the parameters:
        #   - Rename parameters to fit the labels as expected by the widget
        #   - Check parameters to ensure their values are correct, and set widget properties, to use these parameters
        #     while printing into the Excel spreadsheet --> function 'check_parameters_and_set_properties' to be
        #     overridden by each widget will call both the functions to check parameters and config
        super().__init__(
            str_address=str_address,
            dict_parameters_map=dict_parameters_map,
            dict_config=dict_config,
            dict_raw_parameters=dict_raw_parameters,
            str_sheet_name=str_sheet_name,
            dict_default_format=dict_default_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def check_parameters_and_set_properties(
            self,
            dict_parameters: dict,
            dict_config: dict,
    ):
        """
        Function that checks the parameters that are required by the widget:
        - If all expected parameters are correct, the widget properties are set and the function returns (no value
          returned)
        - Otherwise, an exception is thrown.

        Checking parameters means:
        - Check the parameter exists, in case it is a mandatory one. eg. mandatory array of data to print in a list
        - Check the format is as expected. eg. has to be a DataFrame
        - Check the values are correct. eg. numeric vs alpha, etc.
        - Check coherence with config. eg. columns configured as "to be printed" effectively exist in the DataFrame.

        Setting properties means:
        - Set the widget instance properties from the parameters/ config
        - So these properties can later be used while writing into Excel.

        Properties for widget_cover_sheet are:
        - str_narrative
        - lst_tokens

        :param dict_parameters: the parameters provided to the module
        :param dict_config: the configuration of the module
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We check the configuration
        try:
            self._str_narrative = dict_config[
                self.const_str__config_label__str_narrative
            ]
        except KeyError:
            raise Exception(f"Configuration '{self.const_str__config_label__str_narrative}' could not be found, which "
                            f"is not expected.")

        # We check the parameters
        try:
            self._lst_tokens = dict_parameters[
                self.const_str__parameter_label__lst_tokens
            ]
        except KeyError:
            raise Exception(f"Parameter '{self.const_str__parameter_label__lst_tokens}' could not be found, which "
                            f"is not expected.")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def write_to_excel(
            self,
            wb_workbook: Workbook,
            ws_worksheet: Worksheet,
    ):
        """
        Function that writes the widget into the excel spreadsheet. When called, the widget is fully initialized,
        with all its properties set.

        Note: it is while writing into excel that we transcode formats, from "format config" as dictionary to "format"
        object as expected by xlsxwriter.

        :param ws_worksheet: the worksheet to print the widget
        :param wb_workbook:
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        ws_worksheet.write(
            xl_cell_to_rowcol(self.str_address)[0],
            xl_cell_to_rowcol(self.str_address)[1],
            self.str_narrative.format(*self.lst_tokens)
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import logging
from inspect import stack

import click
from pandas import DataFrame, read_csv, read_excel

from com.enovation.helper.click.python_literal_argument_and_option import PythonLiteralOption

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('df-load-csv')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='csv')
@click.option('-c', '--columns', type=list, cls=PythonLiteralOption, default=[],
              help="Columns labels to load, others will be discarded.")
def df_load_csv(ctx_context, file, columns, alias):
    """
    Load listed columns from a csv file into a dataframe that is labelled using an alias for later use.

    :param alias: name given to the dataframe being loaded, so it can be used later on by other commands.
    :param columns: columns from the CSV file to be loaded. Other columns will be ignored.
    :param ctx_context: Click context.
    :param file: csv file to load
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'csv' file '{file}'."
        )

    # We load the csv
    df_the_measures: DataFrame = read_csv(file)
    _logger.info(f"CSV file read, shape '{df_the_measures.shape}'.")

    # We only keep the listed columns, if such "listed columns" were even provided as parameter
    if columns:

        # We check the listed columns are effectively in the data source
        lst_the_inexistant_columns: list = list(set(columns) - set(df_the_measures.columns))
        if len(lst_the_inexistant_columns) > 0:
            raise Exception(
                f"selected columns '{', '.join(lst_the_inexistant_columns)}' are not among existing ones "
                f"'{', '.join(df_the_measures.columns)}'."
            )

        # We filter out non listed columns
        df_the_measures = df_the_measures[list(columns)]
        _logger.info(f"Dataframe reduced, shape '{df_the_measures.shape}'.")

    # We update the context data store
    ctx_context.obj[alias] = df_the_measures
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_csv"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('df-load-xls')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='xls')
@click.option('-c', '--columns', type=list, cls=PythonLiteralOption, default=[],
              help="Columns labels to load, others will be discarded.")
def df_load_xls(ctx_context, file, columns, alias):
    """
    Load listed columns from a xls-x file into a dataframe that is labelled using an alias for later use.

    :param alias: name given to the dataframe being loaded, so it can be used later on by other commands.
    :param columns: columns from the XLS-X file to be loaded. Other columns will be ignored.
    :param ctx_context: Click context.
    :param file: xls-x file to load
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'xls-x' file '{file}'."
        )

    # We load the csv
    df_the_measures: DataFrame = read_excel(file)
    _logger.info(f"XLS-x file read, shape '{df_the_measures.shape}'.")

    # We only keep the listed columns, if such "listed columns" were even provided as parameter
    if columns:

        # We check the listed columns are effectively in the data source
        lst_the_inexistant_columns: list = list(set(columns) - set(df_the_measures.columns))
        if len(lst_the_inexistant_columns) > 0:
            raise Exception(
                f"selected columns '{', '.join(lst_the_inexistant_columns)}' are not among existing ones "
                f"'{', '.join(df_the_measures.columns)}'."
            )

        # We filter out non listed columns
        df_the_measures = df_the_measures[list(columns)]
        _logger.info(f"Dataframe reduced, shape '{df_the_measures.shape}'.")

    # We update the context data store
    ctx_context.obj[alias] = df_the_measures
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_xls"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from datetime import datetime
from inspect import stack
from logging import Logger, getLogger
from os.path import isfile
from pathlib import Path
from typing import Optional, Type

from pandas import ExcelWriter
from xlsxwriter import Workbook
from xlsxwriter.format import Format
from xlsxwriter.worksheet import Worksheet

from com.enovation.toolbox.excel_dashboard.config_checker import ConfigChecker
from com.enovation.toolbox.excel_dashboard.widgets.widget import Widget


class ExcelDashboarder:
    """
    Class that dynamically produce advanced Excel Dashboards.
    """

    const_str__sub_directory_for_widgets: str = "widgets"

    const_str__config_label__str_worksheet_hide: str = "worksheet.hide"
    const_str__config_label__str_widgets: str = "widgets"
    const_str__config_label__str_format: str = "format"
    const_str__config_label__str_sheets: str = "sheets"

    const_str__config_label__str_address: str = "address"
    const_str__config_label__str_parameters_map: str = "parameters_map"
    const_str__config_label__str_config: str = "config"

    const_str__config_label__str_vba: str = "vba"
    const_str__config_label__str_add_vba_project: str = "workbook.add_vba_project"
    const_str__config_label__str_wb_set_vba_name: str = "workbook.set_vba_name"
    const_str__config_label__str_ws_set_vba_name: str = "worksheet.set_vba_name"

    const_str__config_label__str_ws_freeze_panes: str = "worksheet.freeze_panes"

    const_str__config_label__str_wb_set_properties: str = "workbook.set_properties"
    const_str__config_label__str_wb_options: str = "workbook.options"

    const_str__config_label__str_ws_set_column: str = "worksheet.set_column"
    const_str__config_label__ws_set_column__cell_format__as_dict: str = "cell_format"
    const_str__config_label__ws_set_column__first_col__as_int: str = "first_col"
    const_str__config_label__ws_set_column__last_col__as_int: str = "last_col"
    const_str__config_label__ws_set_column__options__as_dict: str = "options"
    const_str__config_label__ws_set_column__options__hidden__as_int: str = "hidden"
    const_str__config_label__ws_set_column__options__level__as_int: str = "level"
    const_str__config_label__ws_set_column__width__as_float: str = "width"

    const_str__config_label__str_ws_set_row: str = "worksheet.set_row"
    const_str__config_label__ws_set_row__cell_format__as_dict: str = "cell_format"
    const_str__config_label__ws_set_row__row__as_int: str = "row"
    const_str__config_label__ws_set_row__options__as_dict: str = "options"
    const_str__config_label__ws_set_row__options__hidden__as_int: str = "hidden"
    const_str__config_label__ws_set_row__options__level__as_int: str = "level"
    const_str__config_label__ws_set_row__height__as_float: str = "height"

    _logger: Logger = getLogger(__name__)

    # This is the repository of all the registered widgets:
    # - The ones proposed by com.enovation packaging
    # - Along with the custom ones that are provided while instantiating the ExcelDashboarder.
    # It is structured as a dictionary:
    # - Key: the widget id (aka "$id" from the json schema)
    # - Value: the widget class
    _dict_widgets_repository: dict[str, Type[Widget]] = None

    # That is the config checker. It contains all json formats:
    # - From low level formats, listed in the class ConfigChecker. Illustration: xlsxwriter functions
    # - To com.enovation widgets, registered in the __init__, and expected to be found in the 'widgets' sub-directory.
    #   Illustration: WidgetTable
    # - To custom widgets, that are to be registered while instantiating the ExcelDashboarder
    _obj_config_checker: ConfigChecker = None

    def __init__(
            self,
            lst_custom_widgets: list[Type[Widget]] = None
    ):
        """
        Subscribe all the widgets, and instantiate the configuration checker.
        This list of widgets can be extended through the parameter lst_widgets.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        Widget.reinitialize_default_formats()

        if lst_custom_widgets is None:
            lst_custom_widgets = []

        # We consolidate the full list of widgets:
        # - The default ones provided by com.enovation, and to be dynamically found under sub-directory "widgets"
        # - Along with the custom ones, provided through the parameter "lst_custom_widgets"
        self._dict_widgets_repository = self._init_widgets_repository(
            lst_custom_widgets=lst_custom_widgets
        )

        # We instantiate the config checker
        self._obj_config_checker = ConfigChecker(
            lst_widgets_json_schemas=[
                i_widget.get_dict_config_schema() for i_widget in self._dict_widgets_repository.values()
            ],
            str_base_json_schema_id="https://enovation.com/excel_dashboard"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _init_widgets_repository(
            self,
            lst_custom_widgets: list[Type[Widget]]
    ) -> dict[str, Type[Widget]]:
        """
        Function that builds the full list of widgets that can be later used by the ExcelDashboarder:
        - It dynamically screen the sub-directory "widgets", and search for the default widgets (aka the ones from
          com.enovation)
        - And it appends the custom widgets, if any.

        These widgets are not checked, so we might later end up discovering they are not properly configured/
        implemented (eg. no json schema file...).

        Two steps to discover and load the default widgets:
        - For a given directory path, we recursively load all python modules (aka files.py)
          https://stackoverflow.com/questions/3365740/how-to-import-all-submodules
        - Once all pythons modules are loaded, we can get all the widgets, as python classes deriving from Widget
          https://stackoverflow.com/questions/3862310/how-to-find-all-the-subclasses-of-a-class-given-its-name

        :param lst_custom_widgets: the custom widgets, to add on top of the default com.enovation ones
        :return: a list of class that extend class Widget
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We recursively import python modules under the sub-directory "widgets"
        # This is implemented in __init__.py, under the same package as this class

        # We initialize the repository of default widgets
        _dict_the_repository_of_widgets_classes: dict[str, Type[Widget]] = self._initialize_default_widgets_repository(
            cls_super_class=Widget
        )

        self._logger.debug(
            f"The ExcelDashboard application recognized '{len(_dict_the_repository_of_widgets_classes)}' widgets: "
            f"{', '.join(_dict_the_repository_of_widgets_classes)}"
        )

        # Eventually, we append the custom widgets provided: nothing to be done, as python classes for the widgets were
        # previously loaded, so they were found from the above _initialize_default_widgets_repository.
        # We will just check they are effectively available in the list
        _missing_widgets: list[str] = [
            i_widget.get_widget_id() for i_widget in lst_custom_widgets
            if i_widget.get_widget_id() not in _dict_the_repository_of_widgets_classes
        ]
        if len(_missing_widgets) > 0:
            self._logger.error(f"Some custom widgets could not be registered... they will be discarded: "
                               f"{', '.join(_missing_widgets)}")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning a dictionary "
                           f"containing '{len(_dict_the_repository_of_widgets_classes)}' records.")
        return _dict_the_repository_of_widgets_classes

    def _initialize_default_widgets_repository(
            self,
            cls_super_class: Type[Widget]
    ) -> dict[str, Type[Widget]]:
        """
        Function that registers all widgets, aka python classes that extend the "Widget" class, directly or indirectly.
        Illustration: class A extends class B that extends class C that ... extends class Z that extends "Widget"
        - Classes A to Y indirectly extend "Widget"
        - Class Z extend directly "Widget"
        - All classes A to Z will be returned
        :return: the full list of widget classes, as a dictionary: $id -> widget class
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_the_return: dict[str, Type[Widget]] = {
            i_class.get_widget_id(): i_class for i_class in cls_super_class.__subclasses__()
        }

        for i_sub_class in cls_super_class.__subclasses__():
            _dict_the_return |= self._initialize_default_widgets_repository(cls_super_class=i_sub_class)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return _dict_the_return

    def excelize(
            self,
            p_output_file_path: Path,
            dict_config: dict,
            **dict_raw_parameters
    ):
        """
        :param p_output_file_path:
        :param dict_config:
        :param dict_raw_parameters: the named parameters that are required by the widgets being used
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We check the excel configuration
        self._obj_config_checker.validate(dict_config)

        # We initialize the default formats
        self._initialize_default_formats(dict_config=dict_config)

        # We go through the configuration, consuming the parameters, and instantiating the widgets
        dict_the_widgets: dict[str, Widget] = self._instantiate_widgets(
            dict_repository_of_widgets=self._dict_widgets_repository,
            dict_config=dict_config,
            dict_raw_parameters=dict_raw_parameters
        )

        # We produce the excel spreadsheet
        self._produce_excel(
            p_output_file_path=p_output_file_path,
            dict_widgets=dict_the_widgets,
            dict_config=dict_config
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _initialize_default_formats(
            self,
            dict_config: dict,
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        Widget.set_dict_wb_default_format(dict_config.get(self.const_str__config_label__str_format, {}))

        for k_sheet, v_sheet in dict_config[self.const_str__config_label__str_sheets].items():
            Widget.add_dict_ws_default_format(
                str_sheet_name=k_sheet,
                dict_default_format=v_sheet.get(self.const_str__config_label__str_format, {})
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _instantiate_widgets(
            self,
            dict_repository_of_widgets: dict,
            dict_config: dict,
            dict_raw_parameters: dict
    ) -> dict[str, Widget]:
        """
        Function that check the parameter dict_config provided as an input:
        - If parameter is correct, the function is returning (no value returned)
        - Otherwise, an exception is thrown.

        The logic checks widget by widget the deeper structure of the json config.

        :param dict_repository_of_widgets: the repository of the widgets
        :param dict_config: the widget configuration
        :param dict_raw_parameters: the parameters provided to the module
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        dict_the_return: dict[str, Widget] = {}

        # We loop through each sheet
        for i_sheet_name, i_sheet in dict_config[self.const_str__config_label__str_sheets].items():

            # We loop through each widget
            for i_widget_label, i_widget in i_sheet[self.const_str__config_label__str_widgets].items():

                if i_widget_label in dict_the_return:
                    raise Exception(f"'{i_widget_label}' is defined several times in the configuration file, which is "
                                    f"not expected.")

                # We instantiate, and register, the widget
                # Note: while instantiating, parameters and configuration are checked...
                dict_the_return[i_widget_label] = dict_repository_of_widgets[
                    i_widget[Widget.const_str__config__widget_id]
                ](
                    str_address=i_widget[self.const_str__config_label__str_address],
                    dict_parameters_map=i_widget.get(self.const_str__config_label__str_parameters_map, {}),
                    dict_config=i_widget.get(self.const_str__config_label__str_config, {}),

                    # We only select the expected parameters, that we map to the labels expected by the widget
                    dict_raw_parameters=dict_raw_parameters,

                    str_sheet_name=i_sheet_name,
                    dict_default_format=i_widget.get(self.const_str__config_label__str_format, {})
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return dict_the_return

    def _produce_excel(
            self,
            p_output_file_path: Path,
            dict_widgets: dict,
            dict_config: dict,
    ) -> None:

        # We instantiate the Excel Writer, and get the workbook instance

        # We instantiate the excel
        xls_the_writer: ExcelWriter = ExcelWriter(
            path=p_output_file_path,
            engine='xlsxwriter',
            engine_kwargs={"options": dict_config.get(self.const_str__config_label__str_wb_options, {})}
        )
        wb_the_workbook: Workbook = xls_the_writer.book

        # We initialize the workbook properties
        self._initialize_workbook_properties(
            wb_the_workbook=wb_the_workbook,
            dict_config=dict_config,
        )

        # For each worksheet
        for k_sheet, v_sheet in dict_config[self.const_str__config_label__str_sheets].items():

            # We instantiate the worksheet
            ws_the_worksheet: Worksheet = self._instantiate_worksheet(
                str_worksheet_label=k_sheet,
                dict_worksheet_config=v_sheet,
                wb_workbook=wb_the_workbook
            )

            # We freeze panes
            if self.const_str__config_label__str_ws_freeze_panes in v_sheet:
                ws_the_worksheet.freeze_panes(**v_sheet[self.const_str__config_label__str_ws_freeze_panes])

            # We write the widgets
            for k_widget, v_widget in v_sheet[self.const_str__config_label__str_widgets].items():
                dict_widgets[k_widget].write_to_excel(
                    wb_workbook=wb_the_workbook,
                    ws_worksheet=ws_the_worksheet
                )

        # We handle VBAs, if any
        if self.const_str__config_label__str_vba in dict_config:

            _dict_vba_config: dict = dict_config[self.const_str__config_label__str_vba]

            # We check the VBA file
            if not isfile(_dict_vba_config[self.const_str__config_label__str_add_vba_project]):
                raise Exception(
                    f"The vba source file '{_dict_vba_config[self.const_str__config_label__str_add_vba_project]}' does "
                    f"not exist, which is not expected."
                )

            # If the VBA file is correct, we associate it to the workbook
            wb_the_workbook.add_vba_project(_dict_vba_config[self.const_str__config_label__str_add_vba_project])

            # We check if we have some VBA name to set
            if self.const_str__config_label__str_wb_set_vba_name in _dict_vba_config:
                wb_the_workbook.set_vba_name(_dict_vba_config[self.const_str__config_label__str_wb_set_vba_name])
            for k_ws, v_ws in _dict_vba_config.get(self.const_str__config_label__str_ws_set_vba_name, {}).items():
                _ws: Worksheet = wb_the_workbook.get_worksheet_by_name(k_ws)
                if _ws is None:
                    raise Exception(f"Worksheet '{k_ws}' does not exist, which is not expected.")
                _ws.set_vba_name(v_ws)

        # # Previously, we had to do the below trick, as 2 files were persisted:
        # # - One proper xlsm file
        # # - One buggy xlsx file, that we had to remove...
        # # Eventually, end of Dec 2022, this issue is not faced anymore... So we can get rid of the below trick.
        #     # Close the Pandas Excel writer and output the Excel file.
        #     xls_the_writer.close()
        #
        #     # Bug two files are generated, the expected xlsm and a buggy xlsx that needs to be removed...
        #     _path_xlsx_file: Path = p_output_file_path.with_suffix('.xlsx')
        #     if _path_xlsx_file.is_file():
        #         self._logger.debug(f"We are deleting the xlsx file generated by default...")
        #         os.remove(_path_xlsx_file)
        #     else:
        #         raise Exception(f"We are expecting a buggy xlsx file to be deleted, at path '{_path_xlsx_file}'... "
        #                         f"We could not find it, so we did not delete it!")
        #
        # # No VBA. We simply close the file...
        # else:

        # Close the Pandas Excel writer and output the Excel file.
        xls_the_writer.close()

    def _initialize_workbook_properties(
            self,
            wb_the_workbook: Workbook,
            dict_config: dict,
    ) -> Workbook:
        """
        Instantiate
        :param dict_config:
        :return:
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We set properties
        if self.const_str__config_label__str_wb_set_properties in dict_config:
            dt_the_created_date = dict_config["workbook.set_properties"].get("created", datetime.now())
            if isinstance(dt_the_created_date, str):
                dt_the_created_date = datetime.fromisoformat(dt_the_created_date)
            wb_the_workbook.set_properties({
                'title': dict_config["workbook.set_properties"].get("title", "com.enovation - default title"),
                'subject': dict_config["workbook.set_properties"].get("subject", "com.enovation - default subject"),
                'author': dict_config["workbook.set_properties"].get("author", "com.enovation - default author"),
                'manager': dict_config["workbook.set_properties"].get("manager", "com.enovation - default manager"),
                'company': dict_config["workbook.set_properties"].get("company", "com.enovation - default company"),
                'category': dict_config["workbook.set_properties"].get("category", "com.enovation - default category"),
                'keywords': dict_config["workbook.set_properties"].get("keywords", "com.enovation default keywords"),
                'created': dt_the_created_date,
                'comments': dict_config["workbook.set_properties"].get("comments", "com.enovation - default comments"),
            })
        else:
            wb_the_workbook.set_properties({
                'title': "com.enovation - default title",
                'subject': "com.enovation - default subject",
                'author': "com.enovation - default author",
                'manager': "com.enovation - default manager",
                'company': "com.enovation - default company",
                'category': "com.enovation - default category",
                'keywords': "com.enovation default keywords",
                'created': datetime.now(),
                'comments': "com.enovation - default comments",
            })

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return wb_the_workbook

    def _instantiate_worksheet(
            self,
            str_worksheet_label: str,
            dict_worksheet_config: dict,
            wb_workbook: Workbook
    ) -> Worksheet:

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We instantiate the worksheet
        ws_the_worksheet: Worksheet = wb_workbook.add_worksheet(name=str_worksheet_label)
        ws_the_worksheet.hide_gridlines(option=2)

        ws_the_worksheet.outline_settings(
            symbols_below=False,
            symbols_right=False
        )

        # for each "worksheet.set_column"
        if self.const_str__config_label__str_ws_set_column in dict_worksheet_config:
            for i_set_column in dict_worksheet_config[self.const_str__config_label__str_ws_set_column]:

                # We instantiate the cell format, if any was provided
                dict_the_format: dict = i_set_column.get(
                    self.const_str__config_label__ws_set_column__cell_format__as_dict, None
                )
                obj_the_format: Optional[Format] = None
                if dict_the_format is not None:
                    obj_the_format = wb_workbook.add_format(dict_the_format)

                # We set the column(s)...
                ws_the_worksheet.set_column(
                    first_col=i_set_column[self.const_str__config_label__ws_set_column__first_col__as_int],
                    last_col=i_set_column[self.const_str__config_label__ws_set_column__last_col__as_int],
                    width=i_set_column.get(self.const_str__config_label__ws_set_column__width__as_float, None),
                    cell_format=obj_the_format,
                    options=i_set_column.get(self.const_str__config_label__ws_set_column__options__as_dict, None),
                )

        # for each "worksheet.set_row"
        if self.const_str__config_label__str_ws_set_row in dict_worksheet_config:
            for i_set_row in dict_worksheet_config[self.const_str__config_label__str_ws_set_row]:

                # We instantiate the cell format, if any was provided
                dict_the_format: dict = i_set_row.get(
                    self.const_str__config_label__ws_set_row__cell_format__as_dict, None
                )
                obj_the_format: Optional[Format] = None
                if dict_the_format is not None:
                    obj_the_format = wb_workbook.add_format(dict_the_format)

                # We set the row...
                ws_the_worksheet.set_row(
                    row=i_set_row[self.const_str__config_label__ws_set_row__row__as_int],
                    height=i_set_row.get(self.const_str__config_label__ws_set_row__height__as_float, None),
                    cell_format=obj_the_format,
                    options=i_set_row.get(self.const_str__config_label__ws_set_row__options__as_dict, None),
                )

        # We check if the sheet is to be hidden
        if dict_worksheet_config.get(self.const_str__config_label__str_worksheet_hide, False):
            ws_the_worksheet.hide()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return ws_the_worksheet



from inspect import stack
from logging import Logger, getLogger

from dash.html import Div

from com.enovation.toolbox.dash.widget import Widget


class Sheet:
    """
    The sheet corresponds to one screen in the dash application. It contains:
    - A label: used in the menu in the dash application navigation bar
    - A list of widgets: which are displayed whener the user open the sheet
    """
    _logger: Logger = getLogger(__name__)

    @property
    def lst_widgets(self) -> list[Widget]:
        return self._lst_widgets

    @property
    def str_label(self) -> str:
        return self._str_label

    @property
    def dict_properties(self) -> dict:
        return self._dict_properties

    def __init__(
            self,
            lst_widgets: list[Widget] = None,
            str_label: str = None,
            dict_properties: dict = None,
            *args,
            **kwargs
    ):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
        self._lst_widgets: list[Widget] = lst_widgets
        self._str_label: str = str_label
        self._dict_properties: dict =dict_properties
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")


class SheetAsDiv(Sheet, Div):

    def __init__(
            self,
            lst_widgets: list[Widget] = None,
            str_label: str = None,
            dict_properties: dict = None,
            *args,
            **kwargs
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        Sheet.__init__(
            self,
            lst_widgets=lst_widgets,
            str_label=str_label,
            dict_properties=dict_properties
        )

        if "children" in kwargs:
            raise Exception(f"Unexpected 'children' in kwargs...")

        Div.__init__(
            self,
            children=lst_widgets,
            *args,
            **kwargs
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import logging

import click

from com.enovation.enov import enov


class EnovGroup(click.Group):
    """
    The com.enovation framework comes with a default click group and commands. Later, this click application can be
    extended with client specific commands.

    To merge both universes (aka group enov and commands, and client specific commands), we use this base class, to
    instantiate the client specific click application.
    """
    _logger: logging.Logger = logging.getLogger(__name__)

    def __init__(
            self,
            str_name: str,
            lst_commands: list[click.Command] = None,
    ):
        super(EnovGroup, self).__init__(
            name=str_name,
            callback=enov.callback,
            params=enov.params,
            chain=enov.chain,
            help=enov.help,
            commands=enov.commands
        )

        if lst_commands is not None:
            self.add_commands(lst_commands)

    def add_commands(self, lst_commands: list[click.Command]):
        for i_cmd in lst_commands:
            self.add_command(i_cmd)



import json
from datetime import datetime

import dash
from dash._callback_context import CallbackContext
from dash.exceptions import PreventUpdate
from plotly.subplots import make_subplots

from inspect import stack
from logging import Logger, getLogger

from dash import callback, Output, Input
from plotly.graph_objs import Figure, Scatter
import com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_wid as wid
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.dp_grapher_session import Session
from com.enovation.toolbox.predictability.dp_date_predictability.dp_grapher.tab_p2_explore.tab_explore__data_to_print \
    import DataToGraph, DataToTable


@callback(
    Output(wid.str_the_wid__explore__graph__as_graph, 'figure'),
    [
        Output(
            wid.str_the_wid__explore__table__prefix_to_access_columns +
            i_col +
            wid.str_the_wid__explore__table__suffix_to_access_columns,
            'children'
        )
        for i_col in DatePredictabilityComputer.lst__output__statistics_by_key__column_labels
    ],

    Input(wid.str_the_wid__explore__key__as_drop_down, 'value'),
    Input(wid.str_the_wid__explore__graph__as_graph, 'clickData'),
    prevent_initial_call=True
)
def _callback_to_update_graph(
        str_key: str,
        dict_click: dict
) -> list:
    """
    There are 2 events that could result in updating the graph:
    - We select in the drop down a new key: the whole graph is refreshed
    - We zoom into a predictability data point: in the upper graph, we click on the predictability curve. As a result
      this data point is emphasized, and the bottom part is refreshed
    :param str_key:
    :param dict_click:
    :return: the refreshed figure to display
    """

    _logger: Logger = getLogger(__name__)

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    # We first determine which widget did trigger this callback
    str_the_trigger: str = _get_triggerer(log_the_logger=_logger)

    # If the trigger is the drop down "key": str_the_wid__explore__key__as_drop_down
    if str_the_trigger == wid.str_the_wid__explore__key__as_drop_down:

        Session.debug(f"User selected the key '{str_key}'.")

        # We prepare the data to easily build the figure and table to display
        obj_the_data_to_graph: DataToGraph = DataToGraph(
            obj_predictability=Session.get_predictability_bean(),
            str_filter_on_key=str_key,
        )
        fig_the_return: Figure = _instantiate_figure_to_return(
            log_the_logger=_logger,
            obj_the_data_to_graph=obj_the_data_to_graph
        )

        obj_the_data_to_table: DataToTable = DataToTable(
            obj_predictability=Session.get_predictability_bean(),
            str_filter_on_key=str_key,
        )
        lst_the_return: list = _instantiate_table_to_return(
            log_the_logger=_logger,
            obj_the_data_to_table=obj_the_data_to_table
        )

    # Else, the trigger is a click in the graph: str_the_wid__explore__graph__as_graph
    elif str_the_trigger == wid.str_the_wid__explore__graph__as_graph:

        # We prepare the data to easily build the figure and table to display
        obj_the_data_to_graph: DataToGraph = DataToGraph(
            obj_predictability=Session.get_predictability_bean(),
            str_filter_on_key=str_key,
            dt_as_of_date=datetime.strptime(
                dict_click["points"][0]["x"],
                '%Y-%m-%d'
            )
        )
        fig_the_return: Figure = _instantiate_figure_to_return(
            log_the_logger=_logger,
            obj_the_data_to_graph=obj_the_data_to_graph
        )

        obj_the_data_to_table: DataToTable = DataToTable(
            obj_predictability=Session.get_predictability_bean(),
            str_filter_on_key=str_key,
            dt_as_of_date=datetime.strptime(
                dict_click["points"][0]["x"],
                '%Y-%m-%d'
            )
        )
        lst_the_return: list = _instantiate_table_to_return(
            log_the_logger=_logger,
            obj_the_data_to_table=obj_the_data_to_table
        )

        fig_the_return.add_vline(
            x=datetime.strptime(
                dict_click["points"][0]["x"],
                '%Y-%m-%d'
            )
        )

    # Else, the trigger is unknown
    else:
        _logger.error(f"Callback triggerer is '{str_the_trigger}', which is not expected. We return PreventUpdate, as "
                      f"we do not know what to do...")
        raise PreventUpdate()

    return [fig_the_return] + lst_the_return


def _instantiate_table_to_return(
        log_the_logger: Logger,
        obj_the_data_to_table: DataToTable
) -> list[str]:
    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    # Little bit of formatting...

    # Predictability as percentage
    obj_the_data_to_table.dict_values[
        DatePredictabilityComputer.str__statistics__column_label__predictability_last
    ] = "{:.0%}".format(
        obj_the_data_to_table.dict_values[
            DatePredictabilityComputer.str__statistics__column_label__predictability_last
        ]
    )

    # Dates as dd-mmm-yyyy
    for i_date in [
        DatePredictabilityComputer.str__statistics__column_label__date_first,
        DatePredictabilityComputer.str__statistics__column_label__date_last,
        DatePredictabilityComputer.str__statistics__column_label__measure_min,
        DatePredictabilityComputer.str__statistics__column_label__measure_max,
        DatePredictabilityComputer.str__statistics__column_label__measure_first,
        DatePredictabilityComputer.str__statistics__column_label__measure_last,
    ]:
        obj_the_data_to_table.dict_values[i_date] = obj_the_data_to_table.dict_values[i_date].strftime("%d-%b-%Y")

    lst_the_return: list[str] = [
        obj_the_data_to_table.dict_values[i_col]
        for i_col in DatePredictabilityComputer.lst__output__statistics_by_key__column_labels
    ]

    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    return lst_the_return


def _instantiate_figure_to_return(
        log_the_logger: Logger,
        obj_the_data_to_graph: DataToGraph
) -> Figure:
    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    # We instantiate the Figure object that will be rendered
    # It contains two parts:
    # - Row 1, the predictability over time
    # - Row 2, the measures and details leading to compute the predictability
    fig_the_return = make_subplots(
        rows=2,
        cols=1,
        vertical_spacing=0.02,
        shared_xaxes=True
    )

    # Add figure title
    fig_the_return.update_layout(
        title_text=obj_the_data_to_graph.str_figure_title
    )

    # We trace the predictability
    fig_the_return.add_trace(
        Scatter(
            x=obj_the_data_to_graph.s_predictabilities_dates,
            y=obj_the_data_to_graph.s_predictabilities_values,
            name=obj_the_data_to_graph.str_predictabilities_ordinate_label,
        ),
        row=1, col=1
    )

    # We trace the measures
    fig_the_return.add_trace(
        Scatter(
            x=obj_the_data_to_graph.s_measures_dates,
            y=obj_the_data_to_graph.s_measures_values,
            name=obj_the_data_to_graph.str_measures_name,
        ),
        row=2, col=1
    )

    # We trace the cone: the increasing, the decreasing and the final (as the horizontal line)
    fig_the_return.add_trace(
        Scatter(
            x=obj_the_data_to_graph.lst_increasing_cone_dates,
            y=obj_the_data_to_graph.lst_increasing_cone_values,
            name=obj_the_data_to_graph.str_increasing_cone_name,
        ),
        row=2, col=1
    )
    fig_the_return.add_trace(
        Scatter(
            x=obj_the_data_to_graph.lst_decreasing_cone_dates,
            y=obj_the_data_to_graph.lst_decreasing_cone_values,
            name=obj_the_data_to_graph.str_decreasing_cone_name,
        ),
        row=2, col=1
    )
    fig_the_return.add_trace(
        Scatter(
            x=obj_the_data_to_graph.lst_last_measure_dates,
            y=obj_the_data_to_graph.lst_last_measure_values,
            name=obj_the_data_to_graph.str_last_measure_name,
        ),
        row=2, col=1
    )

    # Set axis titles
    fig_the_return.update_xaxes(title_text=obj_the_data_to_graph.str_abscissa_label)
    fig_the_return.update_yaxes(
        title_text=obj_the_data_to_graph.str_predictabilities_ordinate_label,
        row=1, col=1,
        range=[0, 1]
    )
    fig_the_return.update_yaxes(
        title_text=obj_the_data_to_graph.str_measures_ordinate_label,
        row=2, col=1
    )

    fig_the_return.update_traces(mode='lines', hovertemplate=None)
    fig_the_return.update_layout(legend=dict(y=0.5, traceorder='reversed', font_size=16))

    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
    return fig_the_return


def _get_triggerer(log_the_logger: Logger) -> str:
    """
    Function that get from the dash context the widget ID that triggered the callback.

    Assumption: there is one and only trigger trigger.
    - If no trigger (or empty): we return PreventUpdate
    - If more than one trigger: only the first one is processed, and an error message is logged

    :param log_the_logger:
    :return: the widget id that triggered the callback
    """
    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

    ctx_the_dash_context: CallbackContext = dash.callback_context

    str_the_event: str = json.dumps(
        {
            'states': ctx_the_dash_context.states,
            'triggered': ctx_the_dash_context.triggered,
            'inputs': ctx_the_dash_context.inputs
        },
        indent=2
    )

    # If function called without any trigger, this is unexpected. We raise an error, and return "no_update"
    if not ctx_the_dash_context.triggered:
        log_the_logger.error(
            f"Function '{stack()[0].filename} - {stack()[0].function}' called while no trigger.\n{str_the_event}")
        raise PreventUpdate()

    # If the triggered property is not as expected
    if len(ctx_the_dash_context.triggered) < 1:
        log_the_logger.error(
            f"Property 'triggered' contains {len(ctx_the_dash_context.triggered)} record, while one is expected. "
            f"No update could be performed.\n{str_the_event}")
        raise PreventUpdate()
    if len(ctx_the_dash_context.triggered) > 1:
        log_the_logger.error(
            f"Property 'triggered' contains {len(ctx_the_dash_context.triggered)} records, while one is expected. "
            f"Only the first triggered element will be processed.\n{str_the_event}")

    # We get the source of the event
    str_the_trigger = ctx_the_dash_context.triggered[0]['prop_id'].split('.')[0]

    log_the_logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning '{str_the_trigger}'.")

    return str_the_trigger



from datetime import datetime
from inspect import stack
from logging import Logger, getLogger
from typing import Union

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from dash.html import P


class Session:

    _logger: Logger = getLogger(__name__)

    # The Predictability bean to graph, as a global class instance
    _obj_cls_predictability_bean: PredictabilityBean

    # The zoomed Predictability bean to graph, as a global class instance
    # _obj_cls_zoomed_predictability_bean: PredictabilityBean
    # _str_cls_zoomed_key: str
    # _dt_cls_zoomed_date: datetime

    # The logged message, to be printed in the debug
    _lst_cls_debug_samp_children: list[P] = []

    @staticmethod
    def debug(debug_trace: Union[str, list[str]], session_id=None):
        Session._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameter "
                              f"debug_trace of type '{type(debug_trace)}', with value '{debug_trace}'.")

        lst_the_list_of_traces: list[str]
        if isinstance(debug_trace, str):
            lst_the_list_of_traces = [debug_trace]
        else:
            lst_the_list_of_traces = debug_trace

        for i_str in lst_the_list_of_traces:
            Session._lst_cls_debug_samp_children.append(
                P(
                    f">> {debug_trace}",
                    className="m-0 m-0 p-0"
                )
            )
            Session._logger.info(f"{debug_trace}")

        Session._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    @staticmethod
    def get_predictability_bean(session_id=None) -> PredictabilityBean:
        """
        Getter method for the instance of the predictability bean which has been loaded, and kept in the session.
        :param session_id:
        :return:
        """
        return Session._obj_cls_predictability_bean

    @staticmethod
    def set_predictability_bean(obj_predictability_bean: PredictabilityBean, session_id=None):
        """
        Getter method for the instance of the predictability bean which has been loaded, and kept in the session.
        :param obj_predictability_bean:
        :param session_id:
        :return:
        """
        Session._obj_cls_predictability_bean = obj_predictability_bean

    @staticmethod
    def get_debug(session_id=None) -> list[P]:
        """
        Getter method for the debug messages that the application append into the property _lst_cls_debug_samp_children.
        :param session_id:
        :return:
        """
        return Session._lst_cls_debug_samp_children

    # @staticmethod
    # def set_zoom(obj_zoomed_predictability_bean: PredictabilityBean, str_zoomed_key: str, dt_zoomed_date: datetime):
    #     Session._obj_cls_zoomed_predictability_bean = obj_zoomed_predictability_bean
    #     Session._str_cls_zoomed_key = str_zoomed_key
    #     Session._dt_cls_zoomed_date = dt_zoomed_date
    #
    # @staticmethod
    # def get_zoom() -> tuple[PredictabilityBean, str, datetime]:
    #     return Session._obj_cls_zoomed_predictability_bean, Session._str_cls_zoomed_key, Session._dt_cls_zoomed_date
    #
    # @staticmethod
    # def reset_zoom():
    #     Session._obj_cls_zoomed_predictability_bean = None
    #     Session._str_cls_zoomed_key = None
    #     Session._dt_cls_zoomed_date = None
    #
    # @staticmethod
    # def is_zoom():
    #     if Session._obj_cls_zoomed_predictability_bean is None:
    #         return False
    #     return True



from inspect import stack
from logging import Logger, getLogger
from typing import Optional

from pandas import DataFrame
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from com.enovation.toolbox.excel_dashboard.widgets.widget import Widget


class WidgetTable(Widget):
    """
    Widget that prints a table in the Excel Spreadsheet.

    Configuration:
    - The options when calling the function worksheet.add_table. See website for further details
      https://xlsxwriter.readthedocs.io/working_with_tables.html

    Parameters:
    - "df_data": the data, as a dataframe, to print into the table.
    """

    # The labels to parse the json configuration file
    const_str__config_label__str_columns: str = "columns"
    const_str__config_label__str_options: str = "options"
    const_str__config_label__str_conditional_formattings: str = "conditional_formattings"
    const_str__config_label__str_formula: str = "formula"
    const_str__config_label__str_multi_range: str = "multi_range"
    const_str__config_label__str_header_format: str = "header_format"
    const_str__config_label__str_format: str = "format"
    const_str__config_label__str_header: str = "header"
    const_str__config_label__str_range: str = "range"
    const_str__config_label__str_criteria: str = "criteria"
    const_str__config_label__str_type: str = "type"
    const_str__config_label__str_total_row: str = "total_row"

    # The labels for the parameters expected by the widget
    const_str__parameter_label__df_data: DataFrame = "df_data"

    _logger: Logger = getLogger(__name__)

    @property
    def dict_options(self) -> dict:
        return self._dict_options

    @property
    def df_data(self) -> DataFrame:
        return self._df_data

    def __init__(
            self,
            str_address: str,
            dict_parameters_map: dict,
            dict_config: dict,
            dict_raw_parameters: dict,
            str_sheet_name: str,
            dict_default_format: dict
    ):
        """
        Initialize the widget:
        - Calling the Widget super init function
        - Declare the widget properties:
          - Each has their "property" getters
          - They are declared and instantiated to None
          - They are effectively set through the function "check_parameters_and_set_properties"

        :param str_address: the address in a sheet to print the widget
        :param dict_config: the widget configuration (aka the "config" node in the json)
        :param dict_parameters_map: the dictionary to map users parameters to the ones expected by the widget:
                                    - key: the expected label
                                    - value: the actual label
        :param dict_raw_parameters: the parameters as labelled and provided by the users
        :param str_sheet_name: the name of the sheet in which the widget is instantiated (required to handle default
            formats inheritance, from the workbook to the workshoeet to the widget)
        :param dict_default_format: the default format set at the level of the widget. If such default format is not
            configured, we expect an empty dictionary
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We declare the widget properties, initialized to None
        self._dict_options: Optional[dict] = None
        self._df_data: Optional[DataFrame] = None

        # The super init function initializes the widget:
        # - Step 1, we set the widget properties (from nodes from the json configuration): address, parameters map,
        #   config
        # - Step 2, we process the parameters:
        #   - Rename parameters to fit the labels as expected by the widget
        #   - Check parameters to ensure their values are correct, and set widget properties, to use these parameters
        #     while printing into the Excel spreadsheet --> function 'check_parameters_and_set_properties' to be
        #     overridden by each widget will call both the functions to check parameters and config
        super().__init__(
            str_address=str_address,
            dict_parameters_map=dict_parameters_map,
            dict_config=dict_config,
            dict_raw_parameters=dict_raw_parameters,
            str_sheet_name=str_sheet_name,
            dict_default_format=dict_default_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def check_parameters_and_set_properties(
            self,
            dict_parameters: dict,
            dict_config: dict,
    ):
        """
        Function that checks the parameters that are required by the widget:
        - If all expected parameters are correct, the widget properties are set and the function returns (no value
          returned)
        - Otherwise, an exception is thrown.

        Checking parameters means:
        - Check the parameter exists, in case it is a mandatory one. eg. mandatory array of data to print in a list
        - Check the format is as expected. eg. has to be a DataFrame
        - Check the values are correct. eg. numeric vs alpha, etc.
        - Check coherence with config. eg. columns configured as "to be printed" effectively exist in the DataFrame.

        Setting properties means:
        - Set the widget instance properties from the parameters/ config
        - So these properties can later be used while writing into Excel.

        Properties for widget_table are:
        - dict_options
        - df_data

        :param dict_parameters: the parameters provided to the module
        :param dict_config: the configuration of the module
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We check the configuration
        self._dict_options = dict_config[self.const_str__config_label__str_options]

        # We check the parameters
        try:
            self._df_data = dict_parameters[
                self.const_str__parameter_label__df_data
            ]
        except KeyError:
            raise Exception(f"Parameter '{self.const_str__parameter_label__df_data}' could not be found, which "
                            f"is not expected.")

        # We check we have the proper number of columns:
        # - The columns in the dataframe on one side
        # - The "non-formula" columns configured in the table on the other side.
        _lst_non_formula_columns: list = [
            i_col[self.const_str__config_label__str_header]
            for i_col in self.dict_options[self.const_str__config_label__str_columns]
            if self.const_str__config_label__str_formula not in i_col
        ]
        if len(self.df_data.columns) != len(_lst_non_formula_columns):
            raise Exception(f"We do not have a coherent number of columns, as [A] should equal [C]:"
                            f"\n\t- [A] df_data contains '{len(self.df_data.columns)}' records"
                            f"\n\t- [B] configuration contains in total '"
                            f"{len(self.dict_options[self.const_str__config_label__str_columns])}' columns"
                            f"\n\t- [C] among which, '{len(_lst_non_formula_columns)}' are non formula.")

        # Addendum as of 02-Jan-2023: not unit tested!
        # We check columns have unique label...
        _lst_all_columns: list = [
            i_col[self.const_str__config_label__str_header]
            for i_col in self.dict_options[self.const_str__config_label__str_columns]
        ]
        if len(_lst_all_columns) != len(set(_lst_all_columns)):
            # We have duplicated columns! that is not expected...
            for i_col in set(_lst_all_columns):
                _lst_all_columns.remove(i_col)
            raise Exception(
                f"In the configuration of the widget Table, you mentioned several columns with the same labels... "
                f"Check columns labelled '{', '.join(_lst_all_columns)}'."
            )

        # We set the conditional formattings
        self.set_conditional_formattings(
            lst_conditional_formattings=dict_config.get(
                self.const_str__config_label__str_conditional_formattings, []
            ),
            str_reference_address=self.str_address,
            str_row_range_limit="2:"+str(len(self.df_data.index)+1)  # +1 as row range should be 2:2 when one line
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def write_to_excel(
            self,
            wb_workbook: Workbook,
            ws_worksheet: Worksheet,
    ):
        """
        Function that writes the widget into the excel spreadsheet. When called, the widget is fully initialized,
        with all its properties set.

        Note: it is while writing into excel that we transcode formats, from "format config" as dictionary to "format"
        object as expected by xlsxwriter.

        :param ws_worksheet: the worksheet to print the widget
        :param wb_workbook:
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # For each and every "formula" columns:
        # - 1. We add an empty columns in the df_data dataframe that will be printed
        #      This is required, or the whole dataframe is printed, and formula columns in between 2 columns would
        #      be overridden, and not appear
        # - 2 and 3. We instantiate the "format objects" from the "config properties"
        _data = self.df_data.copy()
        _index: int = 0
        if self.const_str__config_label__str_columns in self.dict_options:
            for i_column in self.dict_options[self.const_str__config_label__str_columns]:

                # 1. is there a formula defined?
                if self.const_str__config_label__str_formula in i_column:
                    # noinspection PyTypeChecker
                    _data.insert(_index, column="jsg", value=None, allow_duplicates=True)

                _index += 1

                # 2. is there "header_format" defined?
                if self.const_str__config_label__str_header_format in i_column:
                    i_column[self.const_str__config_label__str_header_format] = self.get_format(
                        dict_format=i_column[self.const_str__config_label__str_header_format],
                        wb_workbook=wb_workbook
                    )

                # 3. is there "format" defined?
                if self.const_str__config_label__str_format in i_column:
                    i_column[self.const_str__config_label__str_format] = self.get_format(
                        dict_format=i_column[self.const_str__config_label__str_format],
                        wb_workbook=wb_workbook
                    )
                else:
                    i_column[self.const_str__config_label__str_format] = self.get_format(
                        dict_format={},
                        wb_workbook=wb_workbook
                    )

        _options: dict = self._dict_options.copy()
        _options["data"] = _data.values.tolist()

        (_first_row, _first_col) = self.tpl_address

        # If there is a total row, the table should be one line longer compared to the number of data rows
        if _options[self.const_str__config_label__str_total_row]:
            _last_row = _first_row + len(self.df_data.index) + 1
        else:
            _last_row = _first_row + len(self.df_data.index)
        _last_col = _first_col + len(self.dict_options[self.const_str__config_label__str_columns]) - 1

        ws_worksheet.add_table(
            first_row=_first_row,
            first_col=_first_col,
            last_row=_last_row,
            last_col=_last_col,
            options=_options
        )

        # We eventually have to override all the formula columns... That is required in case "format" are defined, as
        # they somehow override the formulas...
        _i_col: int = _first_col
        if self.const_str__config_label__str_columns in self.dict_options:
            for i_column in self.dict_options[self.const_str__config_label__str_columns]:

                # is there a formula defined?
                if self.const_str__config_label__str_formula in i_column:

                    # We will print the same formula in each and every row
                    for _i_row in range(len(self.df_data.index)):
                        ws_worksheet.write_formula(
                            row=_first_row + _i_row + 1,
                            col=_i_col,
                            formula=i_column[self.const_str__config_label__str_formula],
                            cell_format=i_column.get(self.const_str__config_label__str_format, None),
                        )

                _i_col += 1

        # We eventually write the conditional formattings
        self.write_conditional_formattings(
            wb_workbook=wb_workbook,
            ws_worksheet=ws_worksheet
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
